from brainbox.deciders import BoilerplateOnDemand
from brainbox.utils.compatibility import *

if __name__ == '__main__':
    controller = BoilerplateOnDemand.Controller()
    #resolve_dependencies(controller)
    controller.install()
    controller.self_test()


from dataclasses import dataclass
from pathlib import Path
from ....framework import ConnectionSettings


@dataclass
class ComfyUIFile:
    url: str
    models_subfolder: str|None = None
    filename: str|None = None
    custom_subfolder: str|None = None

    def get_filename(self) -> str:
        if self.filename is not None:
            return self.filename
        return self.url.split('/')[-1]

@dataclass
class ComfyUIExtension:
    git_url: str
    commit: str|None
    has_requirements: bool = False


@dataclass
class ComfyUISettings:
    connection = ConnectionSettings(20301)

    models_to_download = (
        ComfyUIFile(
            'https://civitai.com/api/download/models/46137?type=Model&format=SafeTensor&size=pruned&fp=fp16',
            'checkpoints',
            'meinamix_meinaV9.safetensors'
        ),
        ComfyUIFile(
            'https://civitai.com/api/download/models/106565?type=Model&format=SafeTensor',
            'loras',
            'cat_lora.safetensors',
        ),
        ComfyUIFile(
            'https://huggingface.co/Kim2091/AnimeSharp/resolve/main/4x-AnimeSharp.pth',
            'upscale_models',
        ),
        ComfyUIFile(
            'https://huggingface.co/SmilingWolf/wd-v1-4-moat-tagger-v2/resolve/main/model.onnx',
            filename='wd-v1-4-moat-tagger-v2.onnx',
            custom_subfolder = 'custom_nodes/ComfyUI-WD14-Tagger/models',
        ),
        ComfyUIFile(
            'https://huggingface.co/SmilingWolf/wd-v1-4-moat-tagger-v2/resolve/main/selected_tags.csv',
            filename='wd-v1-4-moat-tagger-v2.csv',
            custom_subfolder = 'custom_nodes/ComfyUI-WD14-Tagger/models',
        ),
    )

    extensions: tuple[ComfyUIExtension,...] = (
        ComfyUIExtension(
            'https://github.com/pythongosssss/ComfyUI-WD14-Tagger',
            'd33501765c5bf3dca6e90e0ebaa962890999fbc5',
            True
        ),
        ComfyUIExtension(
            'https://github.com/pythongosssss/ComfyUI-Custom-Scripts',
            'ae052b625fb4c0b2dfc1649ae4c2b789f4b2bece'
        ),
        ComfyUIExtension(
            'https://github.com/rgthree/rgthree-comfy',
            '8df5e33109bf4642bc0d720646e32ca2520c1d91'
        )
    )

    extension_requirements: tuple[str,...] = (
        'onnxruntime==1.19.2',
    )




from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, FileIO, INotebookableController, IModelDownloadingController, DownloadableModel,
    File
)
from .settings import VoskSettings
from .model import VoskModel
from pathlib import Path


class VoskController(
    DockerWebServiceController[VoskSettings],
    INotebookableController,
    IModelDownloadingController
):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.9.21',
            ('ffmpeg',),
            allow_arm64=True
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            publish_ports={self.connection_settings.port:8080},
            mount_resource_folders={
                'models': '/models'
            }
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return VoskSettings()

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return VoskModel

    def create_api(self):
        from .api import Vosk
        return Vosk()

    def post_install(self):
        self.download_models(self.settings.models_to_download)

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import Vosk
        file = File.read(Path(__file__).parent/'files/test_voice.wav')
        result = api.execute(BrainBoxTask.call(Vosk).transcribe(file, 'en'))
        tc.assertEqual(1, len(result))
        tc.assertEqual('one little spark and before you know it the whole world is burning', result[0]['text'])
        tc.assertEqual(13, len(result[0]['result']))
        tc.assertIn('conf', result[0]['result'][0])
        tc.assertIn('start', result[0]['result'][0])
        tc.assertIn('end', result[0]['result'][0])
        tc.assertIn('word', result[0]['result'][0])

        yield (
            TestReport
            .last_call(api)
        )

        api.execute(BrainBoxTask.call(Vosk).transcribe_to_array(file, 'en'))
        yield (
            TestReport
            .last_call(api)
        )




from ....framework import DockerWebServiceApi
import requests
from .settings import BoilerplateServerSettings
from .controller import BoilerplateServerController



class BoilerplateServer(DockerWebServiceApi[BoilerplateServerSettings, BoilerplateServerController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)


    def echo(self, argument: str):
        result = requests.post(
            self.endpoint('/echo'),
            json = dict(
                argument=argument,
            )
        )
        if result.status_code!=200:
            raise ValueError(f"Endpoint /echo returned unexpected status code {result.status_code}\n{result.text}")
        return result.json()

    Settings = BoilerplateServerSettings
    Controller = BoilerplateServerController



from brainbox.deciders import ResembleEnhance
from brainbox.utils.compatibility import *

if __name__ == '__main__':
    installer = ResembleEnhance.Controller()
    #resolve_dependencies(installer)

    installer.install()
    installer.self_test()

    #test_on_arm64(installer)




from ....framework import DockerWebServiceApi, FileLike, ISingleLoadableModelApi
from .settings import WhisperSettings
from .model import WhisperModel
from .controller import WhisperController
import requests
from io import BytesIO
import json



class Whisper(DockerWebServiceApi[WhisperSettings, WhisperController], ISingleLoadableModelApi):
    def __init__(self, address: str = None):
        super().__init__(address)


    def transcribe_json(self,
                        file: FileLike.Type,
                        model: str|None = None,
                        initial_prompt: None|str = None,
                        parameters: None|dict = None
                        ):
        if parameters is None:
            parameters = {}
        if initial_prompt is not None:
            parameters['initial_prompt'] = initial_prompt
        parameters = dict(model = model, parameters = parameters)
        json_file = BytesIO()
        json_file.write(json.dumps(parameters).encode('utf-8'))
        json_file.seek(0)

        with FileLike(file, self.cache_folder) as stream:
            reply = requests.post(
                f'http://{self.address}/transcribe',
                files=(
                    ('file', stream),
                    ('parameters', json_file)
                )
            )
        if reply.status_code!=200:
            raise ValueError(reply.text)
        return reply.json()


    def transcribe(self,
                   file: FileLike.Type,
                   model: str|None = None,
                   initial_prompt: None | str = None,
                   parameters: None | dict = None
                   ):
        result = self.transcribe_json(file, model, initial_prompt, parameters)
        return result.get('text','').strip()


    def load_model(self, model: str):
        reply = requests.post(f'http://{self.address}/load_model', json=dict(model=model))
        if reply.status_code != 200:
            raise ValueError(reply.text)

    def get_loaded_model_name(self) -> str|None:
        return requests.get(self.endpoint('/get_loaded_model')).json()

    def get_loaded_model(self):
        reply = requests.get(f'http://{self.address}/get_loaded_model')
        return reply.json()

    Controller = WhisperController
    Settings = WhisperSettings
    Model = WhisperModel

    @classmethod
    def get_ordering_arguments_sequence(cls) -> tuple[str,...]|None:
        return ('model',)



from brainbox.deciders import OpenVoice
from brainbox.utils.compatibility import *

if __name__ == "__main__":
    controller = OpenVoice.Controller()
    #resolve_dependencies(controller)
    controller.install()
    controller.self_test()

    # test_on_arm64(controller)



import os
from dataclasses import dataclass, field
from pathlib import Path
from .repository import Repository
from yo_fluq import FileIO
from ...deployment import IImageBuilder, IExecutor, Command
from .apt_installation import AptInstallation
from .build_context import BuildContext
from .user import User
from .dependencies import Dependencies
from .custom_dependencies import CustomDependencies

@dataclass
class BrainboxImageBuilder(IImageBuilder):
    root_path: Path
    source_image: str
    apt_install: tuple[str,...]|None = None
    repository: Repository|None = None

    install_requirements_as_root: bool = False
    allow_arm64: bool = False
    build_command: tuple[str,...] | None = None
    custom_apt_installation: tuple[str,...] | None = None
    custom_dependencies: tuple[Dependencies|CustomDependencies,...]|None = None
    keep_dockerfile: bool = False

    Repository = Repository
    Dependencies = Dependencies
    CustomDependencies = CustomDependencies


    @property
    def source_is_python(self):
        return ':' not in self.source_image

    @property
    def context(self) -> BuildContext:
        return BuildContext(self.root_path, None)

    def create_dockerfile(self, exec):
        lines = []
        source = self.source_image
        if self.source_is_python:
            source = f'python:{source}'
        lines.append(f'FROM {source}')

        lines.extend(self._create_platform_check_lines())

        context = BuildContext(self.root_path, exec)

        if self.apt_install is not None:
            lines.extend(AptInstallation(self.apt_install).to_commands(context))

        if self.custom_apt_installation is not None:
            lines.extend(self.custom_apt_installation)

        if not self.install_requirements_as_root:
            lines.extend(User().to_commands(context))

        if self.custom_dependencies is None:
            lines.extend(Dependencies().to_commands(context))
        else:
            for dep in self.custom_dependencies:
                lines.extend(dep.to_commands(context))

        if self.install_requirements_as_root:
            lines.extend(User().to_commands(context))

        lines.append('WORKDIR /home/app')

        if self.repository is not None:
            lines.extend(self.repository.to_commands(context))

        lines.append('COPY app/ /home/app/main')
        lines.append('ENTRYPOINT ["python3","/home/app/main/main.py"]')
        return '\n\n'.join(lines)



    def build_image(self, image_name: str, exec: IExecutor) -> None:
        FileIO.write_text(
            self.create_dockerfile(exec),
            self.root_path/'Dockerfile'
        )
        args = []
        command = ['docker', 'build']
        if self.build_command is not None:
            command = ['docker'] + list(self.build_command)
        exec.execute(
            command + ['-t', image_name] + args + ['.'],
            Command.Options(workdir=self.root_path)
        )
        if not self.keep_dockerfile:
            os.unlink(self.root_path/'Dockerfile')



    def _create_platform_check_lines(self):
        allowed_platforms = ["x86_64", "amd64"]
        if self.allow_arm64:
            allowed_platforms.extend(["aarch64","arm64" ])

        ifs = []
        for platform in allowed_platforms:
            ifs.append(f'[ "$arch" = "{platform}" ]')
            ifs.append(' || ')

        msg = 'Build requires ' + ', '.join(allowed_platforms)+', but platform was $arch'

        return [
            'RUN arch="$(uname -m)"; '+''.join(ifs)+ '{ echo >&2 "' + msg + '"; exit 1; }'
        ]






from ...deployment import LocalExecutor, Command
from ...common import Loc
from yo_fluq import FileIO

class GpuRegistry:
    _are_present = None

    @staticmethod
    def _get_fname():
        return Loc.temp_folder / 'gpus.json'

    @staticmethod
    def _from_file() -> bool:
        fname = GpuRegistry._get_fname()
        if not fname.is_file():
            return False
        try:
            data = FileIO.read_json(fname)
        except:
            return False

        if isinstance(data, dict) and 'has_gpus' in data:
            GpuRegistry._are_present = data['has_gpus']
            return True

        return False

    @staticmethod
    def _from_docker():
        try:
            LocalExecutor().execute(['docker', 'run', '--gpus', 'all', 'hello-world'],
                                    Command.Options(ignore_exit_code=False))
            GpuRegistry._are_present = True
        except:
            GpuRegistry._are_present = False

    @staticmethod
    def are_present():
        if GpuRegistry._are_present is None:
            if not GpuRegistry._from_file():
                GpuRegistry._from_docker()
                FileIO.write_json(dict(has_gpus=GpuRegistry._are_present), GpuRegistry._get_fname())
        return GpuRegistry._are_present


import os
from typing import Iterable
from unittest import TestCase

import requests

from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, INotebookableController, File
)
from ...common import check_if_its_sound, download_file
from .settings import OpenVoiceSettings
from pathlib import Path
import zipfile


class OpenVoiceController(
    DockerWebServiceController[OpenVoiceSettings],
    INotebookableController,
):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.9.21',
            ('ffmpeg',),
            BrainboxImageBuilder.Repository(
                'https://github.com/myshell-ai/OpenVoice.git',
                'bb79fa78a5a7a7a3d7602b7f6d48705213a039c7',
            ),
            allow_arm64=True
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            publish_ports={self.connection_settings.port:8080},
            mount_resource_folders={
                'pretrained/model/checkpoints' : '/home/app/checkpoints',
                'models' : '/models',
                'voices': '/voices',
                'temp': '/temp'
            },
        )

    def post_install(self):
        path = self.resource_folder('pretrained')/'checkpoints_1226.zip'
        download_file(
            'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/checkpoints_1226.zip',
            path
        )
        if not zipfile.is_zipfile(path):
            raise zipfile.BadZipFile(f"The file {path} is not a valid zip file.")

        folder = self.resource_folder('pretrained','model')
        os.makedirs(folder, exist_ok=True)
        with zipfile.ZipFile(path, 'r') as zip_ref:
            zip_ref.extractall(folder)



    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return OpenVoiceSettings()

    def create_api(self):
        from .api import OpenVoice
        return OpenVoice()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import OpenVoice

        reference_speaker = File.read(Path(__file__).parent / "lina.wav")
        api.execute(BrainBoxTask.call(OpenVoice).train('test_voice', [reference_speaker]))
        yield TestReport.last_call(api).href('training').with_comment("Training a model for a reference voice")

        source_speaker = File.read(Path(__file__).parent/"nikita.wav")
        task = BrainBoxTask.call(OpenVoice).generate(source_speaker, 'test_voice')
        result_file = api.execute(task)

        check_if_its_sound(api.open_file(result_file).content, tc)
        yield TestReport.last_call(api).href('href').result_is_file(File.Kind.Audio)





import uuid

from foundation_kaia.marshalling import endpoint
from .interface import IBrainboxService
from ...controllers import ControllerRegistry
from ...job_processing import IPlanner, MainLoop, Core, Job, SimplePlanner, OperatorLogItem, FailedJobArgument
from dataclasses import dataclass, field
from pathlib import Path
from ...common import Loc, Locator
from sqlalchemy import create_engine, select, Engine
from sqlalchemy.orm import Session
from threading import Thread
from datetime import datetime, timedelta
from ...job_processing.core.job import BrainBoxBase
import time

@dataclass
class BrainBoxServiceSettings:
    registry: ControllerRegistry
    planner: IPlanner = field(default_factory=SimplePlanner)
    port: int = 18090
    debug_output: bool = False
    locator: Locator = Loc
    stop_controllers_at_termination: bool = True




class BrainBoxService(IBrainboxService):
    def __init__(self, settings: BrainBoxServiceSettings):
        self.settings = settings


    def run(self):
        self.engine = create_engine('sqlite:///'+str(self.settings.locator.db_path))
        BrainBoxBase.metadata.create_all(self.engine)
        core = Core(self.engine, self.settings.registry, self.settings.locator, self.settings.debug_output)
        self.loop = MainLoop(core, self.settings.planner, self.settings.stop_controllers_at_termination)
        self.loop_thread = Thread(target=self.loop.run)
        self.loop_thread.start()

    @endpoint(url='/jobs/add', method='POST')
    def base_add(self, jobs:list[dict]) -> list[str]:
        now = datetime.now()
        ids = []
        with Session(self.engine) as session:
            for i, job_dict in enumerate(jobs):
                if 'id' not in job_dict:
                    job_dict['id'] = str(uuid.uuid4())
                job = Job(**job_dict)
                job = job.set_defaults(now + timedelta(microseconds=i))
                ids.append(job.id)
                session.add(job)
            session.commit()
        return ids

    @endpoint(url='/jobs/join', method='GET')
    def base_join(self,
                  ids: list[str],
                  time_limit_in_secods: int|None = None,
                  allow_failures: bool = False
                  ) -> list:
        begin = datetime.now()
        while True:
            with Session(self.engine) as session:
                jobs: list[Job] = list(session.scalars(select(Job).where(Job.id.in_(ids))))
            if len(jobs) != len(ids):
                missing = list(set(ids) - set(j.id for j in jobs))
                raise ValueError(f'Join_execute is missing jobs for ids {missing}')
            id_to_job = {}
            for job in jobs:
                if not job.finished:
                    break
                if not job.success:
                    if allow_failures:
                        id_to_job[job.id] = FailedJobArgument(job)
                    else:
                        stars='*~'*30
                        raise ValueError(f"{job.decider}/{job.decider_parameter}:{job.method} threw an error:\n{stars}\n{job.error}\n{stars}")
                else:
                    id_to_job[job.id] = job
            if len(id_to_job) == len(ids):
                result = []
                for id in ids:
                    if isinstance(id_to_job[id], FailedJobArgument):
                        result.append(id_to_job[id])
                    else:
                        result.append(id_to_job[id].result)
                return result
            if time_limit_in_secods is not None:
                delta = datetime.now() - begin
                if delta.total_seconds() > time_limit_in_secods:
                    raise ValueError(f"Could not wait for join in {time_limit_in_secods} seconds")
            time.sleep(0.1)

    @endpoint(url='/jobs/result', method='GET')
    def result(self, id: str):
        with Session(self.engine) as session:
            result = list(session.scalars(select(Job.result).where(Job.finished.and_(Job.id == id))))
            if len(result)==0:
                return None
            return result[0]


    @endpoint(url='/jobs/summary', method='GET')
    def summary(self, ids: list[str]|None = None, batch_id: str|None = None) -> list[dict]:
        query = select(
            Job.id, Job.decider, Job.decider_parameter, Job.method, Job.batch, Job.info, Job.progress,
            Job.received_timestamp, Job.ready, Job.ready_timestamp, Job.assigned, Job.assigned_timestamp,
            Job.accepted, Job.accepted_timestamp, Job.finished, Job.finished_timestamp,
            Job.success, Job.error,
        )
        if ids is not None:
            query = query.where(Job.id.in_(ids))
        if batch_id is not None:
            query = query.where(Job.batch==batch_id)
        with Session(self.engine) as session:
            rows = list(dict(r._mapping) for r in session.execute(query))
        return rows

    @endpoint(url='/jobs/cancel', method='POST')
    def cancel(self, job_id: str|None = None, batch_id: str|None = None):
        self.loop.cancel(job_id, batch_id)

    @endpoint(url='/jobs/job', method='GET')
    def job(self, id: str):
        with Session(self.engine) as session:
            result = list(session.scalars(select(Job).where(Job.finished.and_(Job.id == id))))
        if len(result) == 0:
            raise ValueError(f"Can't find a job with the id {id}")
        return result[0]

    @endpoint(url='/jobs/operator_log', method='GET')
    def get_operator_log(self, entries_count: int = 100) -> list[OperatorLogItem]:
        return self.loop.core.operator_log.log_items[-entries_count:]

    @endpoint(url='/shutdown', method='POST')
    def shutdown(self):
        self.loop.terminate()
        self.loop_thread.join()
        self.engine.dispose()

    @property
    def cache_folder(self) -> Path:
        return self.settings.locator.cache_folder




from .step import IStep, IStepFactory
from .simple_steps import FilterStep, ConstantStep
from .brainbox_steps import BrainBoxMappingStep, PromptMappingStepFactory
from .representation_steps import ProbabilisticEqualRepresentationStep, SortedRepresentationStep


from dataclasses import fields as get_dataclass_fields, Field
import inspect
from pathlib import Path
from .workflow import IWorkflow
from .....framework import FileLike, BrainBoxTask, IOneBrainBoxTaskFactory
from yo_fluq import FileIO


class TemplateWorkflow(IWorkflow, IOneBrainBoxTaskFactory):
    @staticmethod
    def meta(
            field_name: str|None = None,
            node_title: str|None = None,
            maps_to_job_id: bool = False,
            is_input_file: bool = False,
            is_model_with_priority: None|int = None,
            skip_auto_substitution: bool = False,
    ):
        return dict(
            field_name=field_name,
            node_title=node_title,
            maps_to_job_id = maps_to_job_id,
            is_input_file = is_input_file,
            is_model_with_priority = is_model_with_priority,
            skip_auto_substitution = skip_auto_substitution
        )

    def find_id_by_title(self, js: dict, title: str) -> list[str]:
        results = []
        for key, value in js.items():
            if value['_meta']['title'] == title:
                results.append(key)
        return results

    def find_single_node_by_title(self, js: dict, title: str) -> dict:
        keys = self.find_id_by_title(js, title)
        if len(keys) == 0:
            raise ValueError(f"Can't find the node with title {title}")
        if len(keys) > 1:
            raise ValueError(f"Too many nodes with title {title}")
        node = js[keys[0]]
        return node

    def find_ids_with_field(self, js: dict, field) -> list[str]:
        results = []
        for key, value in js.items():
            if field in value['inputs']:
                results.append(key)
        return results

    def make_substitution(self, js: dict, field: Field, value: object):
        field_name = field.metadata.get('field_name', None)
        if field_name is None:
            field_name = field.name
        node_title = field.metadata.get('node_title', None)

        if node_title is not None:
            node = self.find_single_node_by_title(js, node_title)
            if field_name not in node['inputs']:
                raise ValueError(f'Field {field_name} not found in the node with title {node_title}')
        else:
            keys = self.find_ids_with_field(js, field_name)
            if len(keys) == 0:
                raise ValueError(f"Can't find the node with field {field_name}")
            if len(keys) > 1:
                raise ValueError(f"Too many nodes with field {field_name}. Specify the title.")
            node = js[keys[0]]

        node['inputs'][field_name] = value

    def read_json(self):
        file = inspect.getfile(self.__class__)
        json_file = Path(file.replace('.py','.json'))
        if not json_file.is_file():
            raise ValueError(f"Cannot find workflow {json_file}."
                f"Workflow must be in the same folder as .py file defining {self.__class__.__name__} and have the same name")
        try:
            js = FileIO.read_json(json_file)
        except:
            raise ValueError(f"Cannot read json file {json_file}")
        return js

    def after_substitution(self, js):
        pass

    def preprocess_value(self, field_name, value):
        return value

    def create_workflow_json(self, job_id: str):
        js = self.read_json()

        for field in get_dataclass_fields(type(self)):
            if field.metadata.get('skip_auto_substitution'):
                continue
            value = getattr(self, field.name)
            if field.metadata.get('maps_to_job_id', False):
                value = job_id
            value = self.preprocess_value(field.name, value)
            if field.metadata.get('is_input_file', False):
                value = FileLike.get_name(value, True)
            try:
                self.make_substitution(js, field, value)
            except Exception as ex:
                raise ValueError(f"Error when processing field {field.name}") from ex
        self.after_substitution(js)
        return js


    def get_input_files(self) -> list[FileLike.Type]:
        result = []
        for field in get_dataclass_fields(type(self)):
            if field.metadata.get('is_input_file', False):
                result.append(getattr(self, field.name))
        return result

    def to_task(self):
        ordering_token_components = {}
        for field in get_dataclass_fields(type(self)):
            ordering = field.metadata.get('is_input_file', None)
            if ordering is not None:
                ordering_token_components[ordering] = str(getattr(self, field.name))
        ordering_token ='/'.join(ordering_token_components[k] for k in sorted(ordering_token_components))
        arguments = {k:v for k,v in self.__dict__.items() if k!='_id'}
        arguments['workflow_type'] = type(self)

        if not hasattr(self, '_id'):
            self._id = BrainBoxTask.safe_id()

        return BrainBoxTask(
            decider='ComfyUI',
            decider_method='run_on_dictionary',
            arguments=arguments,
            ordering_token=ordering_token,
            id = self._id
        )






from typing import *
from ..app import IBrainBoxTask

def fix_decider_and_decider_method(
        decider:  Union[str,type,Callable],
        decider_method: Union[None, str, Callable]) -> tuple[str,str, Callable]:

    _method_to_validate = None
    fixed_decider_method = None

    if isinstance(decider, type):
        fixed_decider = decider.__name__
    elif isinstance(decider, str):
        fixed_decider = decider
    elif callable(decider):
        if decider_method is not None:
            raise ValueError("decider is callable, but decider_method is not None. Callable decider sets both decider and decider_method")
        _method_to_validate = decider
        parts = decider.__qualname__.split('.')
        if len(parts) > 2:
            raise ValueError("If `decider` is a function, it must be exactly 2 parts in qualname, Type.Method")
        fixed_decider = parts[0]
        fixed_decider_method = parts[1]
    else:
        raise ValueError(f'decider must be str, type, or callable type method, but was {decider}')

    if callable(decider_method):
        fixed_decider_method = decider_method.__name__
        _method_to_validate = decider_method
    elif isinstance(decider_method, str):
        fixed_decider_method = decider_method
    elif decider_method is None:
        pass
    else:
        raise ValueError(f'decider must be str, method or None, but was {decider}')

    return fixed_decider, fixed_decider_method, _method_to_validate


def fix_arguments_and_dependencies(
        arguments,
        dependencies,
        fake_dependencies: Optional[List[Union[str,'IBrainBoxTask']]]) -> tuple[dict, dict]:
    if arguments is None:
        arguments = {}
    if dependencies is None:
        dependencies = {}

    for key in arguments:
        if key in dependencies:
            raise ValueError("Key {key} exist in both arguments and dependencies")

    fixed_dependencies = {}
    for key, value in dependencies.items():
        if isinstance(value, IBrainBoxTask):
            fixed_dependencies[key] = value.get_resulting_id()
        elif isinstance(value, str):
            fixed_dependencies[key] = value
        else:
            raise ValueError(f"Dependencies should be str or BrainBoxTask, but at key {key} was {value}")

    fixed_arguments = {}
    for key, value in arguments.items():
        if isinstance(value, IBrainBoxTask):
            fixed_dependencies[key] = value.get_resulting_id()
        else:
            fixed_arguments[key] = value

    if fake_dependencies is not None:
        for index, fake_dependency in enumerate(fake_dependencies):
            if isinstance(fake_dependency, IBrainBoxTask):
                fake_dependency = fake_dependency.get_resulting_id()
            elif isinstance(fake_dependency, str):
                pass
            else:
                raise ValueError(f"Fake Dependency should be list of str or BrainBoxTask, but element #{index} was {fake_dependency}")
            fixed_dependencies['*fake_dependency_'+str(index)] = fake_dependency

    return fixed_arguments, fixed_dependencies




from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, File, INotebookableController
)
from ...common import VOICEOVER_TEXT, check_if_its_sound
from .settings import TortoiseTTSSettings
from pathlib import Path


class TortoiseTTSController(DockerWebServiceController[TortoiseTTSSettings], INotebookableController):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.9.21',
            repository=BrainboxImageBuilder.Repository(
                'https://github.com/neonbjb/tortoise-tts.git',
                '8a2563ecabe93c4fb626f876dd0c52c966edef2f',
                remove_files = ('setup.py',)
            ),
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port: 8084},
            mount_resource_folders=dict(
                voices='/home/app/repo/tortoise/voices/',
                hf_models='/home/app/.cache/huggingface/hub/',
                tortoise_models='/home/app/.cache/tortoise/models',
                stash='/stash',
            ),
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return TortoiseTTSSettings()

    def create_api(self):
        from .api import TortoiseTTS
        return TortoiseTTS()

    def post_install(self):
        should_install = False
        for model in [
            self.resource_folder('hf_models')/'models--facebook--wav2vec2-large-960h',
            self.resource_folder('hf_models')/'models--jbetker--tacotron-symbols',
            self.resource_folder('hf_models')/'models--jbetker--wav2vec2-large-robust-ft-libritts-voxpopuli',
            self.resource_folder('tortoise_models') /'models--Manmay--tortoise-tts'
        ]:
            if not model.is_dir():
                should_install = True
                break
        if should_install:
            self.run_auxiliary_configuration(self.get_service_run_configuration(None).as_service_worker('--install'))


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import TortoiseTTS

        TortoiseTTS.export_voice('test_voice', [Path(__file__).parent/'test_voice.wav']).execute(api)
        text = VOICEOVER_TEXT
        files = api.execute(BrainBoxTask.call(TortoiseTTS).dub(text=text, voice='test_voice'))
        yield TestReport.last_call(api).href('voiceover').result_is_array_of_files(File.Kind.Audio).with_comment("Voiceover")
        for i, fname in enumerate(files):
            file = api.open_file(fname)
            check_if_its_sound(file.content, tc)





from typing import *
from .image_source import IImageSource
from .. import IExecutor


class DockerIOImageSource(IImageSource):
    def __init__(self,
                 name: str,
                 docker_url: str,
                 login: str,
                 password: str):
        self.name = name
        self.docker_url = docker_url
        self.login = login
        self.password = password
        self.version = 'latest'

    def _auth_command(self) -> list:
        return ['docker', 'login', self.docker_url, '--username', self.login, '--password', self.password]

    def get_image_name(self):
        return self.name

    def get_remote_name(self):
        return f'{self.docker_url}/{self.name}:{self.version}'

    def push(self, executor: IExecutor):
        print(f'Remote name: {self.get_remote_name()}')
        executor.execute(['docker', 'tag', self.name, self.get_remote_name()])
        executor.execute(self._auth_command())
        executor.execute(['docker', 'push', self.get_remote_name()])

    def pull(self, executor: IExecutor):
        executor.execute(self._auth_command())
        executor.execute(['docker', 'pull', self.get_remote_name()])
        executor.execute(['docker', 'tag', self.get_remote_name(), self.name])










from .open_tts import OpenTTS
from .coqui_tts import CoquiTTS
from .tortoise_tts import TortoiseTTS
from .openvoice import OpenVoice
from .piper import Piper
from .piper_training import PiperTraining
from .zonos import Zonos


from typing import Any
import os
import shutil
import traceback
from dataclasses import dataclass
from threading import Thread

from ...common import Loc, FileIO
from foundation_kaia.marshalling import endpoint
from ..controller import TestReport, DockerWebServiceApi, ControllerRegistry, IController, IModelDownloadingController
from .interface import IControllerService, ControllerServiceStatus, InstallationReport, ControllersSetup

from .logging import Log, LoggingLocalExecutor, LogLogger
from yo_fluq import Query

@dataclass
class RunningInstallation:
    report: InstallationReport
    controller: IController
    thread: Thread|None = None

    def run(self):
        try:
            self.controller.install()
        except:
            error =  traceback.format_exc()
            self.report.error = error

    def exited(self):
        is_alive = self.thread.is_alive()
        return not is_alive


@dataclass
class ControllerServerSettings:
    registry: ControllerRegistry
    port: int = 8091
    api: Any = None

class ControllerService(IControllerService):
    def __init__(self, settings: ControllerServerSettings):
        self.settings = settings
        self.running_installation: RunningInstallation|None = None

    def get_controller(self, decider: str|type):
        name = ControllerRegistry.to_controller_name(decider)
        return self.settings.registry.get_controller(name)

    @endpoint(url='/controllers/install', json_pickle_result=True)
    def install(self, decider: str|type, join: bool = True) -> InstallationReport|None:
        if self.running_installation is not None and not self.running_installation.exited:
            raise ValueError("Another installation is in progress")
        self.running_installation = None

        controller = self.get_controller(decider)
        log = Log()
        controller.context._executor = LoggingLocalExecutor(log)
        controller.context._logger = LogLogger(log)

        installation_report = InstallationReport(controller.get_name(), log)
        installation = RunningInstallation(installation_report, controller)
        installation.thread = Thread(target=installation.run)
        installation.thread.start()
        self.running_installation = installation

        if join:
            return self.join_installation()


    @endpoint(url='/controllers/uninstall')
    def uninstall(self, decider: str|type, purge: bool = False):
        self.get_controller(decider).uninstall(purge)


    @endpoint(url='/controllers/run')
    def run(self, decider: str|type, parameter: str | None = None) -> str:
        if parameter == '':
            parameter = None
        return self.get_controller(decider).run(parameter)


    @endpoint(url='/controllers/stop')
    def stop(self, decider: str|type, instance_id: str):
        self.get_controller(decider).stop(instance_id)


    @endpoint(url='/controllers/join-installation', json_pickle_result=True)
    def join_installation(self) -> InstallationReport:
        if self.running_installation is None:
            raise ValueError("No installation is in progress")
        if not self.running_installation.exited():
            self.running_installation.thread.join()

        result = self.running_installation
        self.running_installation = None
        if result.report.error is not None:
            raise ValueError(f"Installation threw an exception\n{result.report.error}")
        return result.report




    @endpoint(url='/controllers/installation-report', json_pickle_result=True)
    def installation_report(self) -> InstallationReport:
        if self.running_installation is None:
            raise ValueError("No installation is in progress")
        return self.running_installation.report



    @endpoint(url='/controllers/status', json_pickle_result=True)
    def status(self):
        installation_statuses = self.settings.registry.get_installation_statuses()
        containers = []
        for name, installation_status in installation_statuses.items():
            if not installation_status.installed or installation_status.dockerless_controller:
                containers.append(ControllerServiceStatus.Controller(name, installation_status))
                continue

            container = self.get_controller(name)
            instances = []
            for instance_id, parameter in container.get_running_instances_id_to_parameter().items():
                address = None
                api = container.find_api(instance_id)
                if isinstance(api, DockerWebServiceApi):
                    address = api.address
                instances.append(ControllerServiceStatus.Instance(
                    instance_id,
                    parameter,
                    address
                ))

            report_path = (self.settings.registry.locator.self_test_path / name)
            has_report =  report_path.is_file()
            has_errors = False
            if has_errors:
                report = FileIO.read_pickle(report_path)
                if report.error is not None:
                    has_errors = True

            containers.append(ControllerServiceStatus.Controller(
                name,
                installation_status,
                has_report,
                has_errors,
                tuple(instances)
            ))

        installation = None
        if self.running_installation is not None and not self.running_installation.exited():
            installation = self.running_installation.report.name

        return ControllerServiceStatus(
            containers,
            installation
        )

    @endpoint(url='/controllers/self_test', json_pickle_result=True)
    def self_test(self, decider: str|type) -> TestReport:
        controller = self.get_controller(decider)
        return controller.self_test(locator=self.settings.registry.locator, api=self.settings.api)


    @endpoint(url='/controllers/delete_self_test')
    def delete_self_test(self, decider: str|type):
        controller = self.get_controller(decider)
        path = Loc.self_test_path / controller.get_name()
        if path.is_file():
            os.unlink(path)


    def _inner_path(self, controller, path: str):
        while path.startswith('/'):
            path = path[1:]
        path = controller.context.resource_folder()/path
        path.relative_to(controller.context.resource_folder_root)
        return path

    @endpoint(url='/resources/list')
    def list_resources(self, decider: str|type, path: str) -> list[str]|None:
        controller = self.get_controller(decider)
        root_path = controller.resource_folder()
        path = self._inner_path(controller, path)
        if not path.is_dir():
            return None
        return (Query
                .folder(path,'**/*')
                .where(lambda z: z.is_file())
                .select(lambda z: str(z.relative_to(root_path)).replace('\\','/'))
                .to_list()
                )

    @endpoint(url='/resources/delete')
    def delete_resource(self, decider: str|type, path: str, ignore_errors: bool = False):
        controller = self.get_controller(decider)
        path = self._inner_path(controller, path)
        if path.is_file():
            os.unlink(path)
        elif path.is_dir():
            shutil.rmtree(path)
        else:
            if not ignore_errors:
                raise ValueError(f"Cannot delete {path} of decider {controller.get_name()}: no such file/folder")

    @endpoint(url="/resources/setup", method='POST')
    def setup(self, setup: ControllersSetup):
        from .setuper import Setuper
        setuper = Setuper(setup, self)
        setuper.make_all()

    @endpoint(url='/resources/download_models', method='POST')
    def download_models(self, decider: str|type, models: list):
        controller = self.get_controller(decider)
        if not isinstance(controller, IModelDownloadingController):
            raise ValueError(f"Decider {decider} is not ModelDownloadingController")
        controller.download_models(models)






from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder,
    BrainBoxApi, BrainBoxTask, File,
    OnDemandDockerController
)
from .settings import VideoToImagesSettings
from pathlib import Path



class VideoToImagesController(OnDemandDockerController[VideoToImagesSettings]):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.11.11',
            allow_arm64=True
        )

    def get_default_settings(self):
        return VideoToImagesSettings()

    def post_install(self):
        if (self.resource_folder('cache')/'huggingface/hub/models--sentence-transformers--clip-ViT-B-32').is_dir():
            return
        self.run_with_configuration(self.get_run_configuration(['--install']))

    def create_api(self):
        from .api import VideoToImages
        return VideoToImages()

    def get_run_configuration(self, arguments: list[str] | None = None):
        if arguments is None:
            arguments = []
        return RunConfiguration(
            None,
            command_line_arguments=arguments,
            detach_and_interactive=False,
            mount_resource_folders={
                'cache' : '/home/app/.cache'
            }
        )


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import VideoToImages

        VideoToImages.upload_video(Path(__file__).parent/'sample.mp4').execute(api)
        result = api.execute(BrainBoxTask.call(VideoToImages).process(VideoToImages.AnalysisSettings(
            'sample.mp4',
            max_produced_frames=5,
        )))
        tc.assertEqual(5, len(result))
        yield TestReport.last_call(api).href('run')


        result = api.execute(BrainBoxTask.call(VideoToImages).process(VideoToImages.AnalysisSettings(
            'sample.mp4',
            max_produced_frames=5,
            add_semantic_comparator=True,
        )))
        tc.assertEqual(5, len(result))
        yield TestReport.last_call(api).href('run-comparator')


        result = api.execute(BrainBoxTask.call(VideoToImages).process(VideoToImages.AnalysisSettings(
            'sample.mp4',
            max_produced_frames=5,
            buffer_by_laplacian_in_ms=500,
        )))
        tc.assertEqual(5, len(result))
        yield TestReport.last_call(api).href('run_bufferer')





from abc import ABC, abstractmethod
from .dto import *
from ..controller import ControllerRegistry, TestReport
from yo_fluq import Query


class IControllerService(ABC):
    @abstractmethod
    def status(self) -> ControllerServiceStatus:
        pass

    @abstractmethod
    def install(self, decider: str|type, join: bool = True) -> InstallationReport|None:
        pass


    @abstractmethod
    def uninstall(self, decider: str|type, purge: bool = False):
        pass

    @abstractmethod
    def run(self, decider: str|type, parameter: str|None = None) -> str:
        pass

    @abstractmethod
    def stop(self, decider: str|type, instance_id: str|None = None):
        pass

    @abstractmethod
    def join_installation(self) -> InstallationReport:
        pass

    @abstractmethod
    def installation_report(self) -> InstallationReport:
        pass

    @abstractmethod
    def self_test(self, decider: str|type) -> TestReport:
        pass

    @abstractmethod
    def delete_self_test(self, decider: str|type):
        pass

    @abstractmethod
    def list_resources(self, decider: str|type, path: str) -> list[str]|None:
        pass

    @abstractmethod
    def delete_resource(self, decider: str|type, path: str, ignore_errors: bool = False):
        pass

    @abstractmethod
    def setup(self, setup: ControllersSetup):
        pass

    @abstractmethod
    def download_models(self, decider: str|type, models: list):
        pass

    def controller_status(self, decider: str|type) -> ControllerServiceStatus.Controller:
        status = self.status()
        name = ControllerRegistry.to_controller_name(decider)
        return Query.en(status.containers).where(lambda z: z.name == name).single()


from brainbox import BrainBox
from brainbox.deciders import HelloBrainBox
from unittest import TestCase

def uploads(test_case: TestCase, api: BrainBox.Api):
    """
    ### Upload files, required by deciders
    
    Some deciders, especially speech-to-text or image analysis, 
    use files as inputs. You may feed them directly in the call:
    """
    from brainbox import File

    file = File('hello.txt', "Hello, world!")
    test_case.assertEqual(file.name, 'hello.txt')
    test_case.assertEqual(file.content, b'Hello, world!')

    length = api.execute(BrainBox.Task.call(HelloBrainBox).file_length(file))
    test_case.assertEqual(length, 13)

    """
    However, that has the potential to overload the database as well, 
    and should be avoided.
    
    If you run BrainBox server at the machine you're running `api`, 
    you may pass the path of the file:

    ```python
    import tempfile
    from pathlib import Path

    path = file.write(tempfile.gettempdir())
    test_case.assertIsInstance(path, Path)

    length = api.execute(BrainBox.Task.call(HelloBrainBox).file_length(path))
    test_case.assertEqual(length, 13)
    ```
    
    This solution is dirty as it won't work with a remote BrainBox, 
    and would force you to rewrite your codebase in case of such change.
    
    To avoid this, you may choose to upload the file to the BrainBox
    cache folder instead:
    """

    api.upload(file.name, file)
    length = api.execute(BrainBox.Task.call(HelloBrainBox).file_length(file.name))
    test_case.assertEqual(length, 13)

    """
    Sometimes, files are required to have specific names, or some
    recoding. In these cases, deciders usually have static methods
    that incapsulate these procedures in Prerequisites:
    """
    upload_prerequisite = HelloBrainBox.file_upload(file)
    upload_prerequisite.execute(api)
    length = api.execute(BrainBox.Task.call(HelloBrainBox).file_length(file.name))
    test_case.assertEqual(length, 13)



from typing import Iterable
from unittest import TestCase

from ....framework import (
    TestReport, BrainboxImageBuilder, IImageBuilder,
    BrainBoxApi, BrainBoxTask,
    OnDemandDockerController
)
from .settings import WhisperXSettings
from pathlib import Path



class WhisperXController(OnDemandDockerController[WhisperXSettings]):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.11.11',
            ('ffmpeg',),
            allow_arm64=False,
            custom_apt_installation=(CUDA_RUNTIME_INSTALLATION,)
        )

    def get_default_settings(self):
        return WhisperXSettings()

    def create_api(self):
        from .api import WhisperX
        return WhisperX()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import WhisperX

        filename = Path(__file__).parent/'files/sample.mp4'
        api.execute(BrainBoxTask.call(WhisperX).execute(filename))
        yield TestReport.last_call(api).href('run')


CUDA_RUNTIME_INSTALLATION = '''
RUN apt-get update && apt-get install -y curl gnupg && \
    mkdir -p /etc/apt/keyrings && \
    curl -fsSL https://developer.download.nvidia.com/compute/cuda/repos/ubuntu2204/x86_64/3bf863cc.pub \
    | gpg --dearmor -o /etc/apt/keyrings/cuda-archive-keyring.gpg && \
    echo "deb [signed-by=/etc/apt/keyrings/cuda-archive-keyring.gpg] https://developer.download.nvidia.com/compute/cuda/repos/ubuntu2204/x86_64 /" \
    > /etc/apt/sources.list.d/cuda.list && \
    apt-get update && \
    apt-get install -y libcudnn8=8.9.4.* cuda-runtime-12-1 && \
    apt-get clean && rm -rf /var/lib/apt/lists/*
'''.strip()


from dataclasses import dataclass
from ..controller import InstallationStatus
from .logging import Log


@dataclass
class ControllerServiceStatus:
    InstallationStatus = InstallationStatus

    @dataclass
    class Instance:
        instance_id: str
        parameter: str | None
        address: str | None

    @dataclass
    class Controller:
        name: str
        installation_status: InstallationStatus
        has_self_test_report: bool = False
        has_errors_in_self_test_report: bool = False
        instances: tuple['ControllerServiceStatus.Instance', ...] = ()


    containers: list['ControllerServiceStatus.Controller']
    currently_installing_container: str|None




@dataclass
class InstallationReport:
    name: str
    log: Log
    error: str|None = None


@dataclass
class ControllerInstanceSetup:
    decider: type|str
    parameter: str|None = None
    loaded_model: str|None = None


@dataclass
class ControllersSetup:
    controllers: tuple[ControllerInstanceSetup,...]

    Instance = ControllerInstanceSetup



from ..core import Core, ICoreAction, OperatorMessage, Job
from sqlalchemy.orm import Session

class TasksStatusUpdater(ICoreAction):
    def apply(self, core: Core):
        updates = []
        for operator_state in core.operator_states.values():
            while not operator_state.results_queue.empty():
                updates.append(operator_state.results_queue.get())

        if len(updates) == 0:
            return

        id_to_job = {}
        with core.new_session() as session:
            for update in updates:
                if update.id not in id_to_job:
                    id_to_job[update.id] = core.get_job_by_id(session, update.id)
                self._apply_update(core, update, id_to_job[update.id])
            session.commit()



    def _apply_update(self, core, update: OperatorMessage, job: Job):
        if update.type == OperatorMessage.Type.accepted:
            job.accepted = True
            job.accepted_timestamp = update.timestamp

        elif update.type == OperatorMessage.Type.error:
            job.error = update.payload
            job.finished = True
            job.success = False
            job.finished_timestamp = update.timestamp

        elif update.type == OperatorMessage.Type.result:
            job.result = update.payload
            job.success = True
            job.finished = True
            job.finished_timestamp = update.timestamp

        elif update.type == OperatorMessage.Type.report_progress:
            job.progress = update.payload

        elif update.type == OperatorMessage.Type.log:
            job.log.append(update.payload)

        else:
            raise ValueError(f"Unknown update type {update.type}")






from .image_source import IImageSource
from .local_image_source import LocalImageSource
from .remote_public_image_source import RemotePublicImageSource
from .docker_io_image_source import DockerIOImageSource
from .hacked_docker_io_image_source import HackedDockerIOImageSource



from .build_context import BuildContext
from yo_fluq import FileIO
from .builder_part import IBuilderPart
from dataclasses import dataclass

@dataclass
class CustomDependencies(IBuilderPart):
    packages: tuple[str, ...]
    no_build_isolation: bool = False
    no_binary: tuple[str,...]|None = None
    index_url: str | None = None


    def to_commands(self, context: BuildContext) -> list[str]:
        packages = ' '.join(self.packages)
        wheel_arguments = []
        if self.no_build_isolation:
            wheel_arguments.append('--no-build-isolation')
        if self.no_binary is not None:
            wheel_arguments.append('--no-binary')
            wheel_arguments.append(','.join(self.no_binary))
        if self.index_url is not None:
            wheel_arguments.append('--index-url')
            wheel_arguments.append(self.index_url)
        wheel_arguments_str = ' '.join(wheel_arguments)

        return [(
            f'RUN pip wheel --no-cache-dir --wheel-dir=/tmp/wheels {wheel_arguments_str} {packages} && '
            f'pip install --no-cache-dir --no-index --find-links=/tmp/wheels {packages} && '
            f'rm -rf /tmp/wheels'
        )]




from brainbox.deciders import Piper
from brainbox.utils.compatibility import *

if __name__ == "__main__":
    controller = Piper.Controller()
    #resolve_dependencies(controller,'3.11')
    #controller.install()
    #controller.self_test()
    test_on_arm64(controller)



from ..executor import IExecutor, Command
from abc import ABC,abstractmethod

class IImageSource(ABC):
    @abstractmethod
    def get_image_name(self):
        pass

    @abstractmethod
    def pull(self, executor: IExecutor):
        pass

    @abstractmethod
    def push(self, executor: IExecutor):
        pass

    @staticmethod
    def image_name_to_container_name(image_name: str):
        image_name = image_name.split('/')[-1]
        image_name = image_name.split(':')[0]
        return image_name

    def get_container_name(self):
        return IImageSource.image_name_to_container_name(self.get_image_name())


    def get_relevant_containers(self, executor: IExecutor) -> tuple[str,...]:
        reply = executor.execute(['docker', 'ps', '-q', '-a', '--filter', f'ancestor={self.get_image_name()}'], Command.Options(return_output=True))
        containers = [z.strip() for z in reply.strip().split('\n') if z.strip()!='']
        return tuple(containers)



    def get_relevant_images(self, executor: IExecutor) -> tuple[str,...]:
        reply = executor.execute(['docker','images',self.get_image_name(),'-q'], Command.Options(return_output=True))
        images = [z.strip() for z in reply.strip().split('\n') if z.strip() != '']
        return tuple(images)




import uuid

import requests
from ....framework import DockerWebServiceApi
from .model import OllamaModel
from .settings import OllamaSettings
from .controller import OllamaController

SESSION_ID = str(uuid.uuid4())

class Ollama(DockerWebServiceApi[OllamaSettings, OllamaController]):
    def __init__(self, address: str|None = None, parameter: str|None = None):
        super().__init__(address, parameter)


    def completions_json(self, prompt: str, **kwargs):
        reply = requests.post(
            f'http://{self.address}/api/generate',
            json=dict(
                model=self.container_parameter,
                prompt=prompt,
                stream=False,
                **kwargs,
            )
        )
        if reply.status_code != 200:
            raise ValueError(f'Status code {reply.status_code}, value\n{reply.text}')
        return reply.json()

    def completions(self, prompt: str, **kwargs):
        return self.completions_json(prompt, **kwargs)['response']

    def question_json(self,
                      prompt: str,
                      system_prompt: str|None = None,
                      options: dict|None = None,
                      num_predict: int|None = None
                      ):
        messages = []
        if system_prompt is not None:
            messages.append(dict(role='system', content=system_prompt))
        messages.append(dict(role='user', content=prompt))
        json = dict(
                model=self.container_parameter,
                messages=messages,
                stream=False,
                session_id = SESSION_ID
            )
        if options is not None:
            json['options'] = options
        if num_predict is not None:
            json['num_predict'] = num_predict
        reply = requests.post(
            f'http://{self.address}/api/chat',
            json=json
        )
        if reply.status_code != 200:
            raise ValueError(f'Status code {reply.status_code}, value\n{reply.text}')
        return reply.json()

    def question(self,
                 prompt: str,
                 system_prompt: str|None = None,
                 options: dict|None = None,
                 num_predict: int | None = None
                 ):
        return self.question_json(prompt, system_prompt, options, num_predict)['message']['content']


    Controller = OllamaController
    Settings = OllamaSettings
    Model = OllamaModel


from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Callable

T = TypeVar("T")

class IParser(ABC, Generic[T]):
    @abstractmethod
    def __call__(self, text: str) -> T:
        pass

    @staticmethod
    def parse_code_block(text):
        buffer = []
        to_buffer = False
        for line in text.split('\n'):
            if line.strip().startswith('```'):
                to_buffer = not to_buffer
                continue
            if to_buffer:
                buffer.append(line)
        return '\n'.join(buffer)


class FunctionalParser(IParser, Generic[T]):
    def __init__(self, function: Callable[[str], T]):
        self.function = function

    def __call__(self, text: str):
        return self.function(text)


from ..interface import InstallationReport
from ....common import HTML


def create_installation_report_page(report: InstallationReport):
    log = report.log
    html = ['<tt>']
    for item in log.items:
        line = f'[{str(item.timestamp):<15}]  {item.data}'
        html.append(line)
        html.append('<br>')
    return ''.join(html)


import uuid
from typing import *
from ....framework import FailedJobArgument, MediaLibrary
from datetime import datetime
from pathlib import Path
import copy

class CollectorToMediaLibraryConvertor:
    def __init__(self, current_job_id: str, cache_folder: Path, tags: dict[str, dict], results: dict):
        self.current_job_id = current_job_id
        self.cache_folder = cache_folder
        self.tags = tags
        self.results = results
        self.errors = []
        self.records = []

    def _parse_result(self, id: str) -> list[str]|None:
        result = self.results[id]
        if isinstance(result, FailedJobArgument):
            self.errors.append(f'Task {id} failed: {result.job.error}')
            return None
        if isinstance(result, str):
            return [result]
        else:
            try:
                return(list(result))
            except:
                raise ValueError(
                    f"To be included in the MediaLibrary, the output of the job must be str or list of strings, but was {result}")


    def _process_one_record(self, id: str, index: int, record: str):
        tags_copy = copy.copy(self.tags[id])

        inline_kwarg = {}

        if (self.cache_folder/record).is_file():
            fname = record
        else:
            fname = str(uuid.uuid4())
            inline_kwarg['inline_content'] = record

        record = MediaLibrary.Record(
            filename=fname,
            holder_location=self.cache_folder,
            timestamp=datetime.now(),
            job_id=id,
            tags=tags_copy,
            **inline_kwarg
        )
        record.tags['option_index'] = index
        self.records.append(record)

    def convert(self):
        for id, tags in self.tags.items():
            if id not in self.results:
                self.errors.append(f'Id {id} not in the list of jobs')
                continue
            results = self._parse_result(id)
            if results is not None:
                for index, record in enumerate(results):
                    self._process_one_record(id, index, record)

        library = MediaLibrary(tuple(self.records), tuple(self.errors))
        fname = f'{self.current_job_id}.output.zip'
        library.save(self.cache_folder / fname)
        return fname


from brainbox.framework.containerized_brainbox import BrainBoxRunner
from brainbox.framework import BrainBoxApi, Loc
from threading import Thread


class ContainerizedBrainboxTestEnvironment:
    def __init__(self, runner: BrainBoxRunner|None = None):
        if runner is None:
            runner = BrainBoxRunner(Loc.root_folder, 8090, True, debug=True)
        self.runner = runner

    def _run_container_thread(self):
        self.runner.get_deployment().run()

    def __enter__(self):
        self.runner.get_deployment().stop().remove().build()
        Thread(target=self._run_container_thread).start()
        api = BrainBoxApi('127.0.0.1:8090')
        api.wait(60)
        return api

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.runner.get_deployment().stop().remove()


from . import Command
from .executor import IExecutor
import subprocess
from .file_system import IFileSystem
from .local_file_system import LocalFileSystem
from .machine import Machine
from .local_executor_scenarios import ExecutionScenario

class LocalExecutor(IExecutor):
    def get_fs(self) -> IFileSystem:
        return LocalFileSystem()

    def get_machine(self) -> Machine:
        return Machine.local()



    def execute_command(self, command: Command):
        kw = {}
        if command.options is not None and command.options.workdir is not None:
            kw['cwd'] = str(command.options.workdir)
        str_cmd = " ".join(command.command)
        scenatio = ExecutionScenario(
            command.command,
            command.options if command.options is not None else Command.Options(),
            str_cmd,
            kw
        )
        return scenatio.execute()




import json
from typing import Iterable
from unittest import TestCase

from ....framework import (
    File, RunConfiguration, TestReport, IImageBuilder, BrainboxImageBuilder,
    DockerWebServiceController, BrainBoxApi, BrainBoxTask, IModelDownloadingController, DownloadableModel
)
from .settings import WhisperSettings, WhisperModel
from pathlib import Path


class WhisperController(DockerWebServiceController[WhisperSettings], IModelDownloadingController):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.9.21',
            ('ffmpeg',),
            allow_arm64=True
        )

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return WhisperModel

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            None,
            publish_ports={self.connection_settings.port:8084},
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return WhisperSettings()

    def create_api(self):
        from .api import Whisper
        return Whisper()

    def run_notebook(self):
        self.run_with_configuration(self.get_service_run_configuration(None).as_notebook_service())

    def post_install(self):
        self.download_models(self.settings.models_to_download)


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import Whisper

        file = File.read(Path(__file__).parent / 'files/test_voice.wav')

        first_time = True
        for model in self.settings.models_to_download:
            result = api.execute(BrainBoxTask.call(Whisper).transcribe_json(file, model.name))
            yield TestReport.last_call(api).with_comment("Speech recognition with Whisper, full output")
            tc.assertEqual(
                'One little spark and before you know it, the whole world is burning.',
                result['text'].strip()
            )

            result = api.execute(BrainBoxTask.call(Whisper).transcribe(file, model.name))
            if first_time:
                yield TestReport.last_call(api).href('recognition').with_comment("Speech recognition with Whisper, text-only output")
            tc.assertEqual(
                'One little spark and before you know it, the whole world is burning.',
                result
            )




from ..core import ICoreAction, Job, Core

class DirtyExitFix(ICoreAction):
    def apply(self, core: Core):
        with core.new_session() as session:
            (
                session
                .query(Job)
                .filter(~Job.finished & Job.assigned)
                .update({
                    Job.assigned: False,
                    Job.assigned_timestamp: None,
                    Job.accepted: False,
                    Job.accepted_timestamp: None,
                    Job.progress: None,
                    Job.log: None
                })
            )
            session.commit()


from dataclasses import dataclass

@dataclass
class DeciderInstanceKey:
    decider_name: str
    parameter: str|None
    def __str__(self):
        return f'{self.decider_name}/{self.parameter}'
    def __hash__(self):
        return self.__str__().__hash__()



from ..core import ICoreAction, Core, Job
from sqlalchemy.orm import Session
from sqlalchemy import select


class RemoveIncorrectJobsAction(ICoreAction):
    def apply(self, core: Core):
        with core.new_session() as session:
            deciders = core.registry.get_deciders_names()
            jobs = list(session.scalars(select(Job).where(~Job.finished & ~Job.decider.in_(deciders))))
            for job in jobs:
                core.close_job(session, job, f'Decider {job.decider} is not found. Available are: {deciders}')
            session.commit()


from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, INotebookableController, BrainBoxExtendedTask, MediaLibrary
)
from .settings import ResemblyzerSettings
from pathlib import Path


class ResemblyzerController(DockerWebServiceController[ResemblyzerSettings], INotebookableController):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.8.20',
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port: 8084}
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return ResemblyzerSettings()

    def create_api(self):
        from .api import Resemblyzer
        return Resemblyzer()


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import Resemblyzer

        model_name = 'test'
        media_library_path = Path(__file__).parent / 'test_media_library.zip'
        prerequisites = Resemblyzer.upload_dataset(model_name, media_library_path, False)


        api.execute(BrainBoxExtendedTask(
            BrainBoxTask.call(Resemblyzer).train(model_name),
            prerequisite=prerequisites
        ))

        yield (TestReport
               .last_call(api)
               .href('train')
               .with_comment("Training the model with pre-uploaded resources")
               )

        media_library = MediaLibrary.read(media_library_path)

        first_time = True
        first_file = None
        for record in media_library.records:
            if record.tags['split'] == 'test':
                file = record.get_file()
                result = api.execute(BrainBoxTask.call(Resemblyzer).classify(file, model_name))
                if first_time:
                    first_file = file
                    yield TestReport.last_call(api).href('inference').with_comment("Running inference with trained model")
                first_time = False
                tc.assertEqual(record.tags['speaker'], result)

        api.execute(BrainBoxTask.call(Resemblyzer).distances(first_file,model_name))
        yield TestReport.last_call(api).href('distances')

        for record in media_library.records:
            file = record.get_file()
            result = api.execute(BrainBoxTask.call(Resemblyzer).vector(file))
            yield TestReport.last_call(api).href('vectorization').with_comment('Getting a vector of a voice file without ')





from ...deployment import IContainerRunner, IExecutor, Command
from dataclasses import dataclass
from .controller_context import ControllerContext
from .run_configuration import RunConfiguration

@dataclass
class BrainBoxRunner(IContainerRunner):
    context: ControllerContext
    configuration: RunConfiguration
    return_output: bool

    def run(self, image_name: str, container_name: str, executor: IExecutor):
        command = self.configuration.generate_command(container_name, image_name, self.context)
        result = executor.execute(command, Command.Options(return_output=self.return_output))
        if result is None:
            return result
        return result.strip()




from abc import ABC, abstractmethod
from foundation_kaia.prompters import Address, AddressBuilder
from copy import deepcopy

class IApplicator(ABC):
    @abstractmethod
    def apply(self, obj, data_to_apply) -> list:
        pass


class SimpleApplicator(IApplicator):
    def __init__(self, address: Address|AddressBuilder|str):
        self.address = Address(address)

    def apply(self, obj, data_to_apply) -> list:
        obj = deepcopy(obj)
        self.address.set(obj, data_to_apply)
        return [obj]

class ListExpansionApplicator(IApplicator):
    def __init__(self, address: Address|AddressBuilder, index_address: Address|AddressBuilder|None = None):
        self.address = Address(address)
        self.index_address = Address(index_address) if index_address is not None else None


    def apply(self, obj, data_to_apply) -> list:
        try:
            data_to_apply = list(data_to_apply)
        except:
            raise ValueError(f"data_to_apply must be convertible to list")
        result = []
        sap = SimpleApplicator(self.address)
        index_sap = None if self.index_address is None else SimpleApplicator(self.index_address)
        for index, item in enumerate(data_to_apply):
            mod = sap.apply(obj, item)
            if index_sap is not None:
                mod = index_sap.apply(mod[0], index)
            result.append(mod[0])
        return result




from brainbox import BrainBox
from brainbox.deciders import HelloBrainBox
from unittest import TestCase

def dependent_tasks(test_case: TestCase, api: BrainBox.Api):
    """
    ### Build a workflow from tasks

    The output of the tasks can serve as an input to other tasks:
    """
    task1 = BrainBox.Task.call(HelloBrainBox).json("Hello")
    task2 = BrainBox.Task.call(HelloBrainBox).json(task1)
    result = api.execute([task1, task2])
    test_case.assertDictEqual(result[0], result[1]['argument'])

    """
    In this case, the `task1` will be executed before `task2` and the output of `task1` will be used
    as an argument for `task2` method.
    """

def collector(test_case: TestCase, api: BrainBox.Api):
    """
    ### Collect the outputs of many tasks

    One particularly important use case of the dependent tasks is when
    several tasks are run, each associated with some __tags__ describing a task,
    and then the outputs of all tasks are collected together and returned as a single entity.

    `Collector` does just this. Declaring collector tasks in a raw form is very cumbersome,
     and we actually use helpers to do this,
     however, it helps to see what's going on under the hood.
    """
    from brainbox.deciders import Collector
    from brainbox import BrainBoxCombinedTask

    id1 = BrainBox.Task.safe_id()
    id2 = BrainBox.Task.safe_id()
    task1 = BrainBox.Task.call(HelloBrainBox).json(0).to_task(id=id1)
    task2 = BrainBox.Task.call(HelloBrainBox).json(1).to_task(id=id2)
    collector = BrainBox.Task.call(Collector).to_array(
        tags = {id1: dict(index=0), id2: dict(index=1)},
        ** {
            id1: task1,
            id2: task2
        }
    )
    pack = BrainBoxCombinedTask(
        resulting_task = collector,
        intermediate_tasks = (task1, task2)
    )
    array = api.execute(pack)
    for item in array:
        test_case.assertIsNone(item['error'])
        test_case.assertEqual(item['tags']['index'], item['result']['argument'])



    """
    Method `BrainBox.Task.safe_id` creates uuid4 and converts it in python literal so it can be used in **kwargs. 
    
    We can shorten this significantly with `Collector.TaskBuilder`:
    """

    builder = Collector.TaskBuilder()
    for i in range(10):
        builder.append(task=BrainBox.Task.call(HelloBrainBox).json(i), tags=dict(index=i))
    pack = builder.to_collector_pack('to_array')
    array = api.execute(pack)
    for item in array:
        test_case.assertIsNone(item['error'])
        test_case.assertEqual(item['tags']['index'], item['result']['argument'])


def media_library(test_case:TestCase, api:BrainBox.Api):
    """
    ### Collect the files from many tasks
    
    One problem remains: if deciders return files, these files won't be collected. 
    To solve it, MediaLibrary structure is used. Essentially it's a zip-file
    that contains all the outputs as well as tags. 
    
    """
    from brainbox import MediaLibrary
    from brainbox.deciders import Collector
    import json

    builder = Collector.TaskBuilder()
    for i in range(10):
        builder.append(task=BrainBox.Task.call(HelloBrainBox).file(i), tags=dict(index=i))
    pack = builder.to_collector_pack('to_media_library')
    path = api.download(api.execute(pack))
    ml = MediaLibrary.read(path)
    for record in ml.records:
        content = record.get_content()
        tags = record.tags
        test_case.assertEqual(tags['index'], json.loads(content)['argument'])







import os
from typing import *
from enum import Enum
from pathlib import Path
from yo_fluq import FileIO
from typing import TypeVar, Generic

TMetadata = TypeVar('TMetadata')

class File(Generic[TMetadata]):
    class Kind(Enum):
        Audio = 1
        Image = 2
        Json = 3
        Text = 4

    def __init__(self,
                 name: str,
                 content: None|bytes|str,
                 kind: Optional['File.Kind'] = None,
                 ):
        self.name = name
        if isinstance(content, str):
            self.content = content.encode('utf-8')
        else:
            self.content = content
        if kind is not None:
            self.kind = kind
        else:
            self.kind = File.guess_kind_from_filename(name)

    @property
    def string_content(self):
        if self.content is None:
            return None
        return self.content.decode('utf-8')

    @staticmethod
    def guess_kind_from_filename(filename: str):
        extension = filename.split('.')[-1]
        kind = None
        for key, value in _FILETYPES.items():
            if extension in value:
                kind = key
        return kind

    @staticmethod
    def read(filename: str|Path):
        filename = Path(filename)
        kind = File.guess_kind_from_filename(filename.name)
        return File(filename.name, FileIO.read_bytes(filename), kind)

    def write(self, folder: Path|str):
        folder = Path(folder)
        os.makedirs(folder, exist_ok=True)
        path = folder/self.name
        FileIO.write_bytes(self.content, path)
        return path


    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f'File({self.name})'

    @property
    def metadata(self) -> TMetadata:
        if not hasattr(self, '_metadata'):
            raise ValueError("Metadata was not set for this file")
        return self._metadata

    @metadata.setter
    def metadata(self, value: TMetadata):
        self._metadata = value

    def has_metadata(self, _type: type = None):
        if not hasattr(self, '_metadata'):
            return False
        if self._metadata is None:
            return False
        if _type is not None and not isinstance(self._metadata, _type):
            return False
        return True





_FILETYPES = {
    File.Kind.Audio: {'wav','mp3','ogg'},
    File.Kind.Image: {'png','jpg','jpeg','bmp'},
    File.Kind.Json: {'json'},
    File.Kind.Text: {'txt'}
}







from resemblyzer import preprocess_wav, VoiceEncoder
from itertools import groupby
from pathlib import Path
from tqdm import tqdm
import numpy as np
import pandas as pd



class WavProcessor:
    def __init__(self):
        self.encoder = VoiceEncoder()

    def get_encoding(self, wav_path: Path):
        preprocessed = preprocess_wav(wav_path)
        encoding = self.encoder.embed_utterance(preprocessed)
        return encoding

    def get_encodings(self, path: Path):
        encodings = {}
        wavs = list(path.glob('**/*.wav'))
        for wav in tqdm(wavs, f'Folder {path.parent.name}', len(wavs)):
            speaker_name = wav.parent.name
            if speaker_name not in encodings:
                encodings[speaker_name] = []
            encodings[speaker_name].append(self.get_encoding(wav))
        return encodings



class Model:
    def __init__(self, embeddings, accuracy = -1, stats = None):
        self.embeddings = embeddings
        self.accuracy = accuracy
        self.stats: None|pd.DataFrame = stats

    def compute_winner(self, embedding):
        winner = None
        winner_score = None

        for key, value in self.embeddings.items():
            score = max(np.inner(embedding, value))
            if winner is None or score > winner_score:
                winner = key
                winner_score = score
        return winner

    def compute_full(self, embedding):
        rows = []
        for key, value in self.embeddings.items():
            for i, score in enumerate(np.inner(embedding, value)):
                rows.append(dict(speaker=key, sample_index=i, score=float(score)))
        return rows

    def evaluate(self, test):
        dfs = []
        total = 0
        matches = 0
        for key, samples in test.items():
            for i, sample in enumerate(samples):
                rows = self.compute_full(sample)
                df = pd.DataFrame(rows)
                df = df.sort_values('score', ascending=False)
                df['order'] = list(range(df.shape[0]))
                df['true_speaker'] = key
                df['source_sample_index'] = i
                dfs.append(df)

                winner = self.compute_winner(sample)
                if winner == key:
                    matches += 1
                total += 1

        if total > 0:
            self.stats = pd.concat(dfs)
            self.accuracy = matches / total
        else:
            self.stats = None
            self.accuracy = None





from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, File, IModelDownloadingController, DownloadableModel
)
from .settings import YoloSettings
from .model import YoloModel
from pathlib import Path


class YoloController(DockerWebServiceController[YoloSettings], IModelDownloadingController):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.11.11',
            allow_arm64=True
        )

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return YoloModel

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port:8084},
        )

    def get_default_settings(self):
        return YoloSettings()

    def create_api(self):
        from .api import Yolo
        return Yolo()

    def post_install(self):
        self.download_models(self.settings.models_to_downloads)


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import Yolo

        api.execute(BrainBoxTask.call(Yolo).analyze(
            File.read(Path(__file__).parent / "test_anime_img.png"),
            "Fuyucchi/yolov8_animeface:yolov8x6_animeface.pt"
        ))

        yield TestReport.last_call(api).href('run')







from server import CoquiInferenceWebApp
from tts_loader import load_tts
import sys
import subprocess
from pathlib import Path
import argparse

def install(model_name, mode):
    print(f"Installing model {model_name} in mode {mode}")
    if mode == 'simple':
        data = load_tts(model_name)
        print(f"Successfully loaded")
        return
    ret_code = subprocess.run(
        [
            sys.executable,
            Path(__file__).parent/'installer_script.py',
            model_name
        ],
        input='y'.encode('ascii'),
    )
    print(f'Installation script returned code {ret_code}')





if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-n','--notebook', action='store_true')
    parser.add_argument('-i','--install')
    parser.add_argument('-m','--install-mode')


    args = parser.parse_args()
    print(f"Running with arguments\n{args}")

    if args.notebook:
        subprocess.call([sys.executable, '-m', 'notebook', '--allow-root', '--port', '8899', '--ip', '0.0.0.0', "--NotebookApp.token=''"], cwd='/repo')
        exit(0)

    if args.install is not None or args.install_mode is not None:
        if args.install is None or args.install_mode is None:
            raise ValueError("`install` and `install-mode` must be set together")
        install(args.install, args.install_mode)
        exit(0)

    factory = CoquiInferenceWebApp()
    factory.create_app().run('0.0.0.0', 8084)



from .cache_manager import ICacheManager
from abc import ABC
from pathlib import Path
import os
from yo_fluq import *

class FileCacheManager(ICacheManager[str], ABC):
    def __init__(self, folder: Path, extension: str):
        self.folder = folder
        os.makedirs(self.folder, exist_ok=True)
        self.extension = extension

    def get_filename(self, key: str) -> Path:
        return self.folder/(key+'.'+self.extension)

    def contains(self, key: str) -> bool:
        return self.get_filename(key).is_file()

    def unsafe_delete(self, key: str):
        os.unlink(self.get_filename(key))


class JsonCacheManager(FileCacheManager):
    def __init__(self, folder: Path):
        super().__init__(folder, 'json')

    def unsafe_get(self, key: str):
        return FileIO.read_json(self.get_filename(key))

    def store(self, key: str, value):
        FileIO.write_json(value, self.get_filename(key))




from typing import *
from .executor import IExecutor, Command
from .file_system import IFileSystem
from .local_executor import LocalExecutor
import os
from pathlib import Path
from .ssh_file_system import SSHFileSystem
from .machine import Machine

class SSHExecutor(IExecutor):
    def __init__(self,  machine: Machine):
        self.inner_executor = LocalExecutor()
        self.machine = machine


    def _get_prefix(self):
        return [
            'ssh',
            f'{self.machine.username}@{self.machine.ip_address}'
        ]

    def fix(self, s: Iterable[str]):
        return [ss.replace('\\','/') for ss in s]


    def execute_command(self, command: Command):
        _command = self._get_prefix() + list(command.command)
        _command = self.fix(_command)
        return self.inner_executor.execute(_command, command.options)

    def get_fs(self) -> IFileSystem:
        return SSHFileSystem(self.machine)

    def get_machine(self) -> Machine:
        return self.machine









from typing import *
import pickle
import json
from pathlib import Path
import os

class FileIO:
    @staticmethod
    def read_bytes(filename):
        with open(filename, 'rb') as file:
            return file.read()

    @staticmethod
    def read_pickle(filename):
        with open(filename,'rb') as file:
            return pickle.load(file)

    @staticmethod
    def read_json(filename):
        with open(filename,'r', encoding='utf-8') as file:
            result = json.load(file)
            return result

    @staticmethod
    def read_text(filename, encoding='utf-8'):
        with open(filename,'r', encoding=encoding) as file:
            return file.read()


    @staticmethod
    def write_pickle(data, filename):
        with open(filename,'wb') as file:
            pickle.dump(data,file)

    @staticmethod
    def write_json(data, filename):
        with open(filename,'w') as file:
            json.dump(data,file,indent=1)

    @staticmethod
    def write_text(data, filename):
        with open(filename,'w', encoding='utf-8') as file:
            file.write(data)


    @staticmethod
    def write_bytes(data, filename):
        with open(filename,'wb') as file:
            file.write(data)

    @staticmethod
    def folder(location: Union[Path, str], pattern: str = '*'):
        if isinstance(location, str):
            location = Path(location)
        elif isinstance(location, Path):
            pass
        else:
            raise ValueError('Location should be either str or Path, but was {0}, {1}'.format(type(location), location))

        if not os.path.isdir(location):
            raise ValueError('{0} is not a directory'.format(location))

        return location.glob(pattern)


from openvoice.api import ToneColorConverter
import torch
from openvoice import se_extractor

class Converter:
    def __init__(self):
        CKPT_CONVERTER = 'checkpoints/converter'
        self.device  = "cuda:0" if torch.cuda.is_available() else "cpu"

        self.tone_color_converter = ToneColorConverter(
            f'{CKPT_CONVERTER}/config.json',
            device=self.device
        )
        self.tone_color_converter.load_ckpt(f'{CKPT_CONVERTER}/checkpoint.pth')

    def convert(self, path):
        source_se, _ = se_extractor.get_se(
            path,
            self.tone_color_converter,
            vad=True
        )
        return source_se






import os
import shutil
from typing import Iterable
from unittest import TestCase
from ....framework import (
    SmallImageBuilder, IImageBuilder, BrainBoxApi, BrainBoxTask, OnDemandDockerController, Loc,
    RunConfiguration, TestReport
)
from .settings import KohyaSSSettings
from pathlib import Path
import zipfile
import time



class KohyaSSController(OnDemandDockerController[KohyaSSSettings]):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            None
        )

    def get_base_configuration(self) -> RunConfiguration:
        return RunConfiguration(
            mount_resource_folders={
                'models' : '/app/models',
                'training': '/app/training'
            },
            detach_and_interactive=False
        )

    def run_gradio(self):
        cfg = self.get_base_configuration()
        cfg.command_line_arguments=['--gradio']
        cfg.publish_ports = {self.settings.port: 7860}
        self.run_with_configuration(cfg)

    def get_default_settings(self):
        return KohyaSSSettings()

    def create_api(self):
        from .api import KohyaSS
        return KohyaSS()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import KohyaSS, LoraTrainingSettings

        yield TestReport.main_section_content(
            "KohyaSS doesn't run the real training in self-test. Self-test checks the input/output structure only"
        )

        lora_settings = LoraTrainingSettings(
            'test_training',
            'Matroskin',
            'cat',
            mock_training=True
        )
        dataset_path = Path(__file__).parent / 'dataset.zip'
        with zipfile.ZipFile(dataset_path, 'w') as zip_file:
            zip_file.writestr('image1.png', '')
            zip_file.writestr('image1.txt', b'cat, fur, whiskers')
        id = api.add(BrainBoxTask.call(KohyaSS).run_lora_training(dataset_path, {}, lora_settings))
        seen_not_none_progress = False
        while True:
            job = api.summary(ids=[id])[0]
            if job['progress'] is not None:
                seen_not_none_progress = True
            time.sleep(0.5)
            if job['finished']:
                break
        result = api.join(id)
        tc.assertTrue(seen_not_none_progress)
        tc.assertTrue((api.cache_folder/result['log_file']).is_file())

        export_folder = Loc.temp_folder/'kohya_ss_self_test'
        shutil.rmtree(export_folder, ignore_errors=True)
        postprocessor = KohyaSS.Exporter(export_folder, False, False)
        postprocessor.postprocess(result, api)

        files = os.listdir(export_folder)
        tc.assertSetEqual(
            {
                result['training_id'] + '.safetensors',
                result['training_id'] + '.safetensors.description.json',
                result['training_id'] + '-000000.safetensors',
                result['training_id'] + '-000000.safetensors.description.json',

            },
            set(files)
        )
        tc.assertFalse(Path(result['training_folder']).is_dir())

        yield TestReport.last_call(api)






DOCKERFILE = f'''
FROM ghcr.io/bmaltais/kohya-ss-gui:latest

COPY kaia_main.py /app/kaia_main.py

ENTRYPOINT ["/usr/local/bin/python","/app/kaia_main.py"]
'''



from .build_context import BuildContext
from yo_fluq import FileIO
from .builder_part import IBuilderPart

class Dependencies(IBuilderPart):
    def to_commands(self, context: BuildContext) -> list[str]:
        result = []
        if context.requirements_preinstall.is_file():
            build_requirements = FileIO.read_text(context.requirements_preinstall).replace('\n', ' ')
            result.append('RUN '+self._create_one_dependency_command(build_requirements))
        result.append(f'ARG REQ={context.requirements_lock.name}')
        result.append(f'COPY {context.requirements_lock.name} /tmp/requirements.txt')
        result.append('RUN '+self._create_one_dependency_command('-r /tmp/requirements.txt'))
        return result


    def _create_one_dependency_command(self, packages):
        return (
            f'pip wheel --no-cache-dir --wheel-dir=/tmp/wheels {packages} && '
            f'pip install --no-cache-dir --no-index --find-links=/tmp/wheels {packages} && '
            f'rm -rf /tmp/wheels'
        )


from .parser import IParser

class _JointParser(IParser[str]):
    def __init__(self, inner_parser: 'BulletPointParser', join_by: str = ', '):
        self.inner_parser = inner_parser
        self.join_by = join_by

    def __call__(self, text: str) -> str:
        return self.join_by.join(self.inner_parser(text))


class BulletPointParser(IParser[list[str]]):
    DEFAULT_BULLET_POINTS = ('• ', '* ', '- ')

    def __init__(self,
                 join: str|None = None,
                 prefixes: tuple[str,...]|None = None
                 ):
        self.join = join
        if prefixes is None:
            prefixes = BulletPointParser.DEFAULT_BULLET_POINTS
        self.prefixes = prefixes

    def __call__(self, text: str):
        result = []
        for line in text.split('\n'):
            for prefix in self.prefixes:
                if line.startswith(prefix):
                    result.append(line[len(prefix):])
                    break
        return result

    def joint(self, by: str =', '):
        return _JointParser(self, by)



from typing import Iterable
from cv2 import cvtColor, COLOR_BGR2GRAY, Laplacian, CV_64F
from PIL import Image
from sentence_transformers import SentenceTransformer, util
from frame import Frame

def get_comparator():
    comparator = SentenceTransformer('clip-ViT-B-32')
    return comparator


def layer_comparator(iterable: Iterable[Frame]):
    old_embedding = None
    cmp = get_comparator()

    for frame in iterable:
        gray_image = cvtColor(frame.frame, COLOR_BGR2GRAY)
        gray_image = Image.fromarray(gray_image)
        embedding = cmp.encode(gray_image, convert_to_tensor=True)

        if old_embedding is not None:
            frame.semantic_comparator_delta = util.cos_sim(embedding, old_embedding).item()

        old_embedding = embedding
        yield frame



import subprocess
from dataclasses import dataclass
from pathlib import Path
import os

class LocalMachine:
    _local = None

    @staticmethod
    def get():
        if LocalMachine._local is not None:
            return LocalMachine._local

        if os.name=='nt':
            user_id = 1000
            group_id = 1000
        else:
            user_id = int(subprocess.check_output(['id', '-u']).decode('ascii').strip())
            group_id = int(subprocess.check_output(['id', '-g']).decode('ascii').strip())
        LocalMachine._local = Machine(user_id, group_id, '127.0.0.1', '')
        return LocalMachine._local


@dataclass
class Machine:
    user_id: int
    group_id: int
    ip_address: str
    username: str


    @staticmethod
    def local():
        return LocalMachine.get()



from .job import Job, BrainBoxBase
from .decider_instance_key import DeciderInstanceKey
from .operator_message import OperatorMessage
from .operator_state import OperatorState, OperatorStateForPlanner
from .job_for_planner import JobForPlanner
from .core import Core, ICoreAction
from .trackable_session_factory import TrackableSessionFactory
from .operator_log import OperatorLog, OperatorLogItem, OperatorLogHandle


from .brainbox_mapping import BrainBoxMapping
from .applicator import SimpleApplicator, ListExpansionApplicator
from .flow import Flow
from .steps import *
from .object_to_task import IObjectConverter, PromptBasedObjectConverter, FunctionalObjectToTask


import subprocess, sys
from pathlib import Path

subprocess.run([
    sys.executable, 
    "-m", "uvicorn", 
    "server:app",
    "--host", "0.0.0.0", 
    "--port", "8080"
], cwd =  Path(__file__).parent)



from typing import Iterable
from ....framework import DockerWebServiceApi, FileLike, File
import requests
from .settings import ZonosSettings
from .controller import ZonosController
import shutil
import os
import numpy as np


class Zonos(DockerWebServiceApi[ZonosSettings, ZonosController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)


    def train(self, speaker: str, samples: Iterable[FileLike.Type]):
        folder = self.controller.resource_folder('voices',speaker)
        shutil.rmtree(folder, ignore_errors=True)
        os.makedirs(folder)
        for sample in samples:
            with open(folder/FileLike.get_name(sample), 'wb') as stream:
                with FileLike(sample, self.cache_folder) as data:
                    stream.write(data.read())
        result = requests.post(
            self.endpoint(f'/train/{speaker}')
        )
        if result.status_code!=200:
            raise ValueError(f"Endpoint train returned error\n{result.text}")

    def voiceover(self,
                  text: str,
                  speaker: str,
                  language='en-us',
                  emotion: list[float|int]|np.ndarray|None = None,
                  speaking_rate: int|float|None = None
                  ):
        if emotion is not None:
            emotion = [float(z) for z in emotion]
        dct = dict(
            text=text,
            speaker=speaker,
            language=language,
            emotion = emotion,
            speaking_rate=speaking_rate
        )
        result = requests.post(
            self.endpoint('/voiceover'),
            json = dct
        )
        if result.status_code!=200:
            raise ValueError(f"Endpoint voiceover returned error\n{result.text}")
        return File(
            self.current_job_id+'.output.wav',
            result.content
        )


    Settings = ZonosSettings
    Controller = ZonosController

    class Emotions:
        Happiness = np.array([1, 0, 0, 0, 0, 0, 0, 0])
        Sadness =   np.array([0, 1, 0, 0, 0, 0, 0, 0])
        Disgust =   np.array([0, 0, 1, 0, 0, 0, 0, 0])
        Fear =      np.array([0, 0, 0, 1, 0, 0, 0, 0])
        Surprise =  np.array([0, 0, 0, 0, 1, 0, 0, 0])
        Anger =     np.array([0, 0, 0, 0, 0, 1, 0, 0])
        Other =     np.array([0, 0, 0, 0, 0, 0, 1, 0])
        Neutral =   np.array([0, 0, 0, 0, 0, 0, 0, 1])




from pathlib import Path

from ....framework import DownloadableModel, IController
from typing import cast

class OllamaModel(DownloadableModel):
    def __init__(self, name: str, location: str|None = None):
        self.name = name
        if location is None:
            location = name
            if '/' not in location:
                location='library/'+location
            if ':' in location:
                location = location.replace(':','/')
            else:
                location = location+'/latest'
        self.location = location

    def get_local_location(self, controller: IController) -> Path:
        return controller.resource_folder('main') / 'models/manifests/registry.ollama.ai/library' / self.location

    @classmethod
    def download(cls, models: list['DownloadableModel'], controller: IController):
        from .controller import OllamaController
        controller = cast(OllamaController, controller)
        for model in models:
            model = cast(OllamaModel, model)
            controller.run_auxiliary_configuration(controller.get_service_run_configuration('').as_service_worker('pull', model.name))


import traceback
import flask
from runner import run_espeak

class EspeakPhonemizerApp:
    def create_app(self):
        app = flask.Flask(type(self).__name__)
        app.add_url_rule('/', view_func=self.index, methods=['GET'])
        app.add_url_rule('/echo', view_func=self.decide, methods = ['POST'])
        return app

    def index(self):
        return "OK"


    def decide(self):
        try:
            data = flask.request.json
            result = run_espeak(data)
            return flask.jsonify(result)
        except:
            return traceback.format_exc(), 500



from brainbox import BrainBox
from brainbox.deciders import HelloBrainBox
from unittest import TestCase

def downloads(test_case: TestCase, api: BrainBox.Api):
    """
    ### Download files, produced by deciders

    Many BrainBox deciders return files with generated images or sounds.
    In order not to overload the BrainBox database with all these
    gigabytes, the files are stored in the cache folder,
    and only files' names are returned.
    """

    filename = api.execute(BrainBox.Task.call(HelloBrainBox).file("Hello, file!"))
    test_case.assertIsInstance(filename, str)

    """
    You may open this file with `api.open_file` method, which returns `File` 
    instance, reading the content of the file on the fly. 
    """

    from brainbox import File
    import json

    file = api.open_file(filename)
    test_case.assertIsInstance(file, File)
    test_case.assertDictEqual(
        {
            'argument': 'Hello, file!',
            'model': 'no_parameter',
            'setting': 'default_setting'
        },
        json.loads(file.content)
    )

    """
    If you don't want to open the file, only download it on the disk:
    """
    from pathlib import Path

    path = api.download(filename)
    test_case.assertIsInstance(path, Path)
    with open(path, 'r') as stream:
        test_case.assertDictEqual(
            {
                'argument': 'Hello, file!',
                'model': 'no_parameter',
                'setting': 'default_setting'
            },
            json.load(stream)
        )

    """
    which simply downloads the file on the disk to the given location
    (by default, in the API's cache folder), and returns the full path to file.
    You may specify the location and the flag to redownload the file even
    if it was already downloaded.
    """



from .template import TemplateWorkflow
from dataclasses import dataclass, field
from .....framework import FileLike, File

@dataclass
class Upscale(TemplateWorkflow):
    input_filename: FileLike.Type = field(metadata=TemplateWorkflow.meta('image','Load Image', is_input_file=True))
    model_name: str = field(default='4x-AnimeSharp.pth', metadata=TemplateWorkflow.meta(is_model_with_priority=0))
    width: int = field(default=1024, metadata=TemplateWorkflow.meta(node_title='Upscale Image'))
    height: int = field(default=1024, metadata=TemplateWorkflow.meta(node_title='Upscale Image'))
    filename_prefix: str = field(default='', metadata=TemplateWorkflow.meta(maps_to_job_id=True))

    def postprocess(self, files: list[File]):
        return files[0]







from uuid import uuid4
from .logger import Logger
from .loc import Loc
from pathlib import Path
from dataclasses import dataclass, field

class DeciderContext:
    def __init__(self):
        self._logger : Logger|None = None
        self._current_job_id: str|None = None
        self._cache_folder: Path|None = None

    @property
    def logger(self) -> Logger:
        if self._logger is None:
            return Logger()
        return self._logger

    @property
    def current_job_id(self) -> str:
        if self._current_job_id is None:
            return str(uuid4())
        return self._current_job_id

    @property
    def cache_folder(self) -> Path:
        if self._cache_folder is None:
            return Loc.cache_folder
        return self._cache_folder


class IDecider:
    @property
    def context(self) -> DeciderContext:
        if not hasattr(self, '_context'):
            self._context = DeciderContext()
        return self._context

    @context.setter
    def context(self, value: DeciderContext):
        self.context = value

    @property
    def cache_folder(self) -> Path:
        return self.context.cache_folder

    @property
    def current_job_id(self) -> str:
        return self.context.current_job_id

    @classmethod
    def get_ordering_arguments_sequence(cls) -> tuple[str,...]|None:
        return None


class ISelfManagingDecider(IDecider):
    def warmup(self, parameter: str|None):
        pass

    def cooldown(self):
        pass




from brainbox.deciders.images.yolo.controller import YoloController
from brainbox.utils.compatibility import *

if __name__ == '__main__':
    controller = YoloController()

    resolve_dependencies(controller)
    test_on_arm64(controller)

    controller.install()
    controller.self_test()



from typing import *
from dataclasses import dataclass
from PIL import Image

@dataclass
class Frame:
    frame: Any
    pil_image: Image.Image
    index: int = 0
    timestamp_in_ms: float = 0
    laplacian: float = 0
    semantic_comparator_delta: float = 0
    simple_comparator_delta: float = 0
    filename: str|None = None



from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, FileIO, INotebookableController, IModelDownloadingController, DownloadableModel,
    File
)
from .settings import WD14TaggerSettings
from .model import WD14TaggerModel
from pathlib import Path
import os

class WD14TaggerController(
    DockerWebServiceController[WD14TaggerSettings],
    IModelDownloadingController,
    INotebookableController
):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.11.11',
            ('ffmpeg', 'libsm6', 'libxext6'),
            BrainboxImageBuilder.Repository(
                "https://github.com/corkborg/wd14-tagger-standalone",
                "f1114c877ed6d1b1311f7e485ec0466d5064eb0c",
            ),
            allow_arm64=True
        )

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return WD14TaggerModel

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        config = RunConfiguration(
            publish_ports={self.settings.connection.port: 8084},
            mount_resource_folders={'models': '/home/app/.cache/huggingface'},
        )
        if self.settings.cpu_share is not None:
            cpu_count = os.cpu_count()
            config.custom_flags=[f'--cpus={self.settings.cpu_share*cpu_count}']
        return config

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return WD14TaggerSettings()

    def create_api(self):
        from .api import WD14Tagger
        return WD14Tagger()

    def post_install(self):
        self.download_models(self.settings.models_to_download)

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import WD14Tagger
        file = File.read(Path(__file__).parent / 'image.png')

        model = self.settings.models_to_download[0].name
        api.execute(BrainBoxTask.call(WD14Tagger).interrogate(image=file, model=model))
        yield TestReport.last_call(api)

        api.execute(BrainBoxTask.call(WD14Tagger).tags(model=model, count=10))
        yield TestReport.last_call(api)





import datetime
import os
import shutil
import time
import zipfile


from pathlib import Path
from ....framework import FileLike, OnDemandDockerApi, FileIO, Loc
from .settings import KohyaSSSettings
from .controller import KohyaSSController
from dataclasses import dataclass
import toml
import subprocess
import re
from enum import Enum
from .exporter import LorasExporter



@dataclass
class LoraTrainingSettings:
    class TrainingType:
        sdxl = 'sdxl_train_network.py'

    character_name: str
    class_name: str
    repeat_count: int = 10
    kohya_ss_executable: str = TrainingType.sdxl
    base_model_name: str = "sdXL_v10VAEFix.safetensors"
    prompt: str|None = None
    mock_training: bool = False
    regularization_repeat_count: int|None = None




class KohyaSS(OnDemandDockerApi[KohyaSSSettings, KohyaSSController]):
    def __init__(self):
        pass

    def _get_dst_folder(self):
        return self.controller.resource_folder('training',self.training_id)

    def _unzip(self, zip_file: FileLike.Type, folder: Path, tmp_file_name: Path):
        with FileLike(zip_file, self.cache_folder) as file:
            FileIO.write_bytes(file.read(), tmp_file_name)
        os.makedirs(folder, exist_ok=True)
        with zipfile.ZipFile(tmp_file_name, 'r') as zip_file:
            for file_name in zip_file.namelist():
                content = zip_file.read(file_name)
                FileIO.write_bytes(
                    content,
                    folder/file_name,
                )
        os.unlink(tmp_file_name)

    def _unzip_dataset(self,
                       dataset_zip_file: FileLike.Type,
                       settings: LoraTrainingSettings
                       ):
        dst = self._get_dst_folder()
        os.makedirs(dst, exist_ok=True)
        image_path = (
                dst
                /'img'
                /f'{settings.repeat_count}_{settings.character_name} {settings.class_name}'
        )
        self._unzip(dataset_zip_file, image_path, dst/'dataset.zip')


    def _unzip_regularization(self,
                              regularization_zip_file: FileLike.Type,
                              settings: LoraTrainingSettings
                              ):
        dst = self._get_dst_folder()
        os.makedirs(dst, exist_ok=True)
        regularization_path = (
            dst
            /'reg'
            /f'{settings.regularization_repeat_count}_{settings.class_name}'
        )
        self._unzip(regularization_zip_file, regularization_path, dst/'regularization.zip')


    def _fix_and_write_toml_file(self, cfg, settings: LoraTrainingSettings, job_id):
        training_path = '/app/training/'+self.training_id+'/'
        models_path = '/app/models/'

        cfg['logging_dir'] = training_path+'log'
        cfg['output_dir'] = training_path+'model'
        cfg['sample_prompts'] = training_path+'model/prompt.txt'
        cfg['train_data_dir'] = training_path+'img'
        cfg['pretrained_model_name_or_path'] = models_path+settings.base_model_name
        cfg['output_name'] = job_id

        if settings.regularization_repeat_count is not None:
            cfg['reg_data_dir'] = training_path+'reg'

        model_folder = self._get_dst_folder()/'model'
        os.makedirs(model_folder, exist_ok=True)
        with open(model_folder/'config.toml', 'w') as file:
            toml.dump(cfg, file)

    def _small_files_and_folders(self, settings: LoraTrainingSettings):
        text = settings.character_name + ", " + settings.class_name
        if settings.prompt is not None:
            text = settings.prompt
        dst_folder = self._get_dst_folder()
        FileIO.write_text(
            text,
            dst_folder / 'model/prompt.txt'
        )
        os.makedirs(dst_folder / 'log', exist_ok=True)

    def _dump_command_line_arguments(self, settings: LoraTrainingSettings):
        arguments = [
            '/home/1000/.local/bin/accelerate',
            'launch',
            '--dynamo_backend',
            'no',
            '--dynamo_mode',
            'default',
            '--mixed_precision',
            'fp16',
            '--num_processes',
            '1',
            '--num_machines',
            '1',
            '--num_cpu_threads_per_process',
            '2',
            f'/app/sd-scripts/{settings.kohya_ss_executable}',
            '--config_file',
            f'/app/training/{self.training_id}/model/config.toml'
        ]
        FileIO.write_json(arguments, self._get_dst_folder()/'parameters.json')


    def _get_logs_file_name(self):
        return  f'{self.training_id}.logs.txt'

    def _monitor(self, s):
        all_percentages = re.findall('steps: +(\d+)%', s)
        if len(all_percentages) > 0:
            try:
                value = int(all_percentages[-1]) / 100
            except:
                value = None
                pass
            self.context.logger.report_progress(value)
        with open(self.cache_folder/self._get_logs_file_name(), 'a') as stream:
            stream.write(s)


    def run_lora_training(self,
                          dataset_zip_file: FileLike.Type,
                          toml_file_content: dict,
                          lora_training_settings: LoraTrainingSettings|dict,
                          regularization_zip_file: FileLike.Type|None = None,
                          description = None
                          ):
        if isinstance(lora_training_settings, dict):
            lora_training_settings = LoraTrainingSettings(**lora_training_settings)

        if (
                (lora_training_settings.regularization_repeat_count is None) !=
                (regularization_zip_file is None)
        ):
            raise ValueError("regularization_repeat_count must be set if and only if regularization_zip_file is provided")

        self.training_id = lora_training_settings.character_name + "_" + self.current_job_id

        dst_folder = self._get_dst_folder()
        shutil.rmtree(dst_folder, ignore_errors=True)

        self._unzip_dataset(dataset_zip_file, lora_training_settings)
        if regularization_zip_file is not None:
            self._unzip_regularization(regularization_zip_file, lora_training_settings)

        self._fix_and_write_toml_file(toml_file_content, lora_training_settings, self.training_id)
        self._small_files_and_folders(lora_training_settings)
        self._dump_command_line_arguments(lora_training_settings)

        config = self.controller.get_base_configuration()
        config.command_line_arguments = ['--train',self.training_id]
        if lora_training_settings.mock_training:
            config.command_line_arguments.append('--mock')
        self.run_container(config, self._monitor)

        error_code_file = dst_folder/'error_code.txt'
        if not error_code_file.is_file():
            raise ValueError(f"No error code was found. The container was probably forcibly removed. Logs are in {self._get_logs_file_name()}")
        error_code=FileIO.read_text(error_code_file)
        if error_code != '0':
            raise ValueError(f"Non-zero error code `{error_code}. Logs are in {self._get_logs_file_name()}")
        description['timestamp'] = str(datetime.datetime.now())
        return dict(
            training_id=self.training_id,
            log_file = self._get_logs_file_name(),
            training_folder = f'training/{self.training_id}',
            description = description,
        )


    Controller = KohyaSSController
    Settings = KohyaSSSettings
    Exporter = LorasExporter
    TrainingSettings = LoraTrainingSettings







from brainbox.deciders import HelloBrainBox
from brainbox.utils.compatibility import resolve_dependencies, test_on_arm64

if __name__ == '__main__':
    controller = HelloBrainBox.Controller()
    #resolve_dependencies(controller)
    controller.install()
    controller.self_test()
    #controller.run_notebook()




import pickle
import traceback
import flask
import os
from recognition import FaceEmbedder
from pathlib import Path
import subprocess
import traceback


class InsightFaceApp:
    def __init__(self, embedder: FaceEmbedder):
        self.embedder = embedder


    def create_app(self):
        app = flask.Flask(type(self).__name__)
        app.add_url_rule('/', view_func=self.index, methods=['GET'])
        app.add_url_rule('/analyze', view_func=self.analyze, methods=['POST'])
        return app

    def index(self):
        return "OK"

    def analyze(self):
        try:
            file = flask.request.files['file']
            data = file.stream.read()
            embeddings = self.embedder.extract_embeddings(data)
            return flask.jsonify(embeddings)
        except:
            return traceback.format_exc(), 500




from dataclasses import dataclass, field
from pathlib import Path
from brainbox.framework.deployment.executor import IExecutor, LocalExecutor

@dataclass
class BuildContext:
    root_path: Path
    executor: IExecutor|None = None

    @property
    def requirements_original(self):
        return self.root_path/'requirements.txt'


    @property
    def requirements_lock(self):
        return self.root_path/'requirements.lock.txt'

    @property
    def requirements_preinstall(self):
        return self.root_path/'requirements.build.txt'

    @property
    def repo_fix_folder(self):
        return self.root_path/'repo_fix'

    @property
    def app_folder(self):
        return self.root_path/'app'


from ....framework import IPrerequisite, BrainBoxApi, Loc, LocalExecutor, ResourcePrerequisite
from pathlib import Path
import os

class TortoiseTTSUploadPrerequisite(IPrerequisite):
    def __init__(self, voice: str, source_file_path: Path):
        self.voice = voice
        self.source_file_path = source_file_path


    def execute(self, api: BrainBoxApi):
        tmp_file = Loc.temp_folder / 'tortoise_tts_export' / (self.source_file_path.name + '.wav')
        os.makedirs(tmp_file.parent, exist_ok=True)
        LocalExecutor().execute(['ffmpeg', '-i', str(self.source_file_path), '-ar', '22050', str(tmp_file), '-y'])
        ResourcePrerequisite('TortoiseTTS', f'voices/{self.voice}/{self.source_file_path.name}.wav', tmp_file).execute(api)
        os.unlink(tmp_file)


import os
from typing import List, Dict, Any, Tuple

import numpy as np
import cv2
from insightface.app import FaceAnalysis


class FaceEmbedder:
    def __init__(
        self,
        model_name: str = "buffalo_l",
        det_size: Tuple[int, int] = (320, 320),
    ):
        """
        Keep a single InsightFace FaceAnalysis instance.
        Always try CUDA first, then fallback to CPU.
        """
        self.model_name = model_name
        self.det_size = det_size
        self.providers = ["CUDAExecutionProvider", "CPUExecutionProvider"]

        self.app = FaceAnalysis(name=self.model_name, providers=self.providers)
        self.app.prepare(ctx_id=0, det_size=self.det_size)

    def extract_embeddings(self, image_bytes: bytes) -> List[Dict[str, Any]]:
        """
        Decode image bytes, detect faces, and return for each face:
          { "bbox": [x1, y1, x2, y2], "embedding": [float, ...] }
        bbox coordinates are rounded to integers.
        """
        arr = np.frombuffer(image_bytes, dtype=np.uint8)
        img = cv2.imdecode(arr, cv2.IMREAD_COLOR)
        if img is None:
            raise ValueError("Failed to decode image from bytes.")

        faces = self.app.get(img)

        results: List[Dict[str, Any]] = []
        for f in faces:
            bbox_xyxy = [
                int(round(f.bbox[0])),
                int(round(f.bbox[1])),
                int(round(f.bbox[2])),
                int(round(f.bbox[3])),
            ]
            emb = f.normed_embedding.astype(float).tolist()  # 512 floats
            results.append({"bbox": bbox_xyxy, "embedding": emb})

        return results


from dataclasses import dataclass

@dataclass
class TrainingSettings:
    language: str = 'en-us'
    sample_rate: int = 22050
    base_model: str = 'lessac.ckpt'
    batch_size: int = 8
    validation_split: float = 0.0
    max_epochs: int = 2164+6
    checkpoint_epochs: int = 1
    precision: int = 32
    num_test_examples: int = 0
    keep_intermediate: bool = False
    single_speaker_to_multi_speaker: bool = False
    continue_existing: bool = False


    



from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, FileIO, INotebookableController, IModelDownloadingController, DownloadableModel
)
from .settings import BoilerplateServerSettings
from pathlib import Path


class BoilerplateServerController(
    DockerWebServiceController[BoilerplateServerSettings],
    INotebookableController,
):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.11.11',
            allow_arm64=True
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            publish_ports={self.connection_settings.port:8080},
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return BoilerplateServerSettings()

    def create_api(self):
        from .api import BoilerplateServer
        return BoilerplateServer()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import BoilerplateServer

        api.execute(BrainBoxTask.call(BoilerplateServer).echo('argument'))


        yield (
            TestReport
            .last_call(api)
            .href('echo')
            .with_comment("Returns JSON with passed arguments and `success` fields")
        )





import os
import pickle
import traceback
import flask
from foundation_kaia.marshalling import Server
from .service import ControllerServerSettings, ControllerService
from flask import Flask
from . import html_helpers
from ...common import Loc
import io
from ..controller import create_self_test_report_page

class ControllerServer(Server):
    def __init__(self, settings: ControllerServerSettings):
        self.settings = settings
        self.service = ControllerService(settings)
        super().__init__(
            settings.port,
            self.service
        )

    def bind_custom_endpoints(self, app):
        app.add_url_rule(
            '/html/controllers/status',
            view_func=self._html_status,
            methods=['GET']
        )

        app.add_url_rule(
            '/html/controllers/installation_report',
            view_func=self._controllers_installation_report,
            methods=['GET']
        )

        app.add_url_rule(
            '/html/controllers/self_test_report/<decider>',
            view_func=self._controllers_self_test_report,
            methods=['GET']
        )

        app.add_url_rule(
            '/resources/upload/<decider_name>/<path:path>',
            view_func=self._resources_upload,
            methods=['POST']
        )

        app.add_url_rule(
            '/resources/download/<decider_name>/<path:path>',
            view_func=self._resources_download,
            methods=['GET']
        )


    def bind_app(self, app: Flask):
        super().bind_app(app)
        self.bind_custom_endpoints(app)
        app.add_url_rule('/', view_func=self._html_status, methods=['GET'])



    def _html_status(self):
        return html_helpers.create_status_page(self.service.status())

    def _controllers_installation_report(self):
        return html_helpers.create_installation_report_page(self.service.installation_report())


    def _controllers_self_test_report(self, decider: str):
        with open(self.settings.registry.locator.self_test_path/decider, 'rb') as file:
            report = pickle.load(file)
        return create_self_test_report_page(report)


    def _resources_upload(self, decider_name: str, path: str):
        try:
            self.service.get_controller(decider_name)
            path = self.settings.registry.locator.resources_folder/decider_name/path
            count = 0
            for key, value in flask.request.files.items():
                if count>0:
                    raise ValueError("Only one file is expected")
                os.makedirs(path.parent, exist_ok=True)
                value.save(path)
                count += 1
            return 'OK'
        except:
            return traceback.format_exc(), 500


    def _resources_download(self, decider_name: str, path: str):
        try:
            self.service.get_controller(decider_name)
            path = self.settings.registry.locator.resources_folder / decider_name / path
            with open(path, 'rb') as file:
                return flask.send_file(
                    io.BytesIO(file.read()),
                    mimetype='application/octet-stream'
                )
        except:
            return traceback.format_exc(), 500



from ..step import IStep
from typing import Any, Callable, Optional
from dataclasses import dataclass
from abc import ABC, abstractmethod
import numpy as np

@dataclass
class RepresentationData:
    item: Any
    category: Any
    category_size: int
    selected: bool


@dataclass
class AbstractRepresentationStep(IStep, ABC):
    count: int
    item_to_category: Callable[[Any], Any]
    item_filter: Optional[Callable[[Any], bool]] = None
    item_to_count: Optional[Callable[[Any], int]] = None

    def _add_item_to_dict(self, item, d):
        category = self.item_to_category(item)
        if self.item_to_count is None:
            delta = 1
        else:
            delta = self.item_to_count(item)
        d[category] = d.get(category, 0) + delta

    def process(self, history: list, current: list) -> list[RepresentationData]:
        categories_count = {}
        data_list = []
        for item in history:
            if self.item_filter is not None and not self.item_filter(item):
                continue
            category = self.item_to_category(item)
            if self.item_to_count is None:
                delta = 1
            else:
                delta = self.item_to_count(item)
            categories_count[category] = categories_count.get(category, 0) + delta

        for item in current:
            category = self.item_to_category(item)
            data_list.append(RepresentationData(
                item,
                category,
                categories_count.get(category,0),
                False
            ))

        if len(current) < self.count:
            for item in data_list:
                item.selected = True
            return data_list

        self.select(data_list)
        return data_list


    def shorten(self, data):
        return [d.item for d in data if d.selected]


    @abstractmethod
    def select(self, data: list[RepresentationData]) -> None:
        pass




from dataclasses import dataclass
from ....framework import ConnectionSettings
from .model import WhisperModel


@dataclass
class WhisperSettings:
    connection: ConnectionSettings = ConnectionSettings(20102)

    models_to_download: tuple[WhisperModel,...] = (
        WhisperModel('base'),
    )





from ...common import SignatureProcessor
from ..app import IBrainBoxTask
from .one_brainbox_task_factory import IOneBrainBoxTaskFactory
from ...common import IDecider



class BrainBoxTaskBuilderResult(IOneBrainBoxTaskFactory):
    def __init__(self,
                 type: type,
                 method:str | None = None,
                 arguments: dict|None = None,
                 ordering_token: str|None = None,
                 fake_dependencies: list[str|IBrainBoxTask]|None = None,
                 decider_parameter: str|None = None,
                 ):
        self._type = type
        self._method = method
        self._arguments: dict|None = arguments
        self._ordering_token = ordering_token
        self._fake_dependencies = fake_dependencies
        self._id = IBrainBoxTask.safe_id()
        self._decider_parameter = decider_parameter


    def to_task(self, id: str|None = None, batch_id: str|None = None, info = None) -> IBrainBoxTask:
        from .task import BrainBoxTask
        actual_id = id if id is not None else self._id

        return BrainBoxTask(
            id = actual_id,
            decider = self._type,
            decider_method = self._method,
            arguments=self._arguments,
            decider_parameter=self._decider_parameter,
            ordering_token=self._ordering_token,
            fake_dependencies=self._fake_dependencies,
            batch=batch_id if batch_id is not None else actual_id,
            info = info,
        )




    def dependent_on(self, *args: str|IBrainBoxTask):
        return BrainBoxTaskBuilderResult(self._type, self._method, self._arguments, self._ordering_token, list(args))


def _get_ordering_token(tp: type[IDecider], arguments: dict):
    ordering_sequence = tp.get_ordering_arguments_sequence()
    if ordering_sequence is None:
        return None
    ordering_sequence = {argname: index for index, argname in enumerate(ordering_sequence)}
    ordering_parts = {}
    for key, value in arguments.items():
        if key in ordering_sequence:
            ordering_parts[ordering_sequence[key]] = str(value)
    for idx in ordering_sequence.values():
        if idx not in ordering_parts:
            ordering_parts[idx] = '*'
    ordering_token = '/'.join(ordering_parts[k] for k in sorted(ordering_parts))
    return ordering_token



class BrainBoxTaskBuilder:
    def __init__(self,
                 _type: type[IDecider],
                 method: str|None = None,
                 parameter: str|None = None,
                 ):
        self._type = _type
        self._method = method
        self._decider_parameter = parameter


    def __call__(self, *args, **kwargs) -> BrainBoxTaskBuilderResult:
        method = self._method
        if method is None:
            method = '__call__'
        method_instance = getattr(self._type, method)
        validator = SignatureProcessor.from_signature(method_instance)
        arguments = validator.to_kwargs_only(*args, **kwargs)
        return BrainBoxTaskBuilderResult(
            self._type,
            self._method,
            arguments,
            _get_ordering_token(self._type, arguments),
            decider_parameter = self._decider_parameter
        )


    def __getattr__(self, method) -> 'BrainBoxTaskBuilder':
        if not hasattr(self._type, method) or not callable(getattr(self._type, method)):
            raise ValueError(f"Type {self._type} doesn't contain method {method}")
        return BrainBoxTaskBuilder(self._type, method, self._decider_parameter)






import traceback

from fastapi import FastAPI, HTTPException
from pydantic import BaseModel
import os 
from pathlib import Path
import subprocess
import uuid
from fastapi.responses import FileResponse

app = FastAPI()

AUDIO_OUTPUT_DIR = Path('/cache')
MODEL_DIR = Path("/models")


class TTSRequest(BaseModel):
    text: str
    model: str
    noise_scale: float|None
    length_scale: float|None
    noise_w: float|None
    speaker: int|None

class ModelDownloadRequest(BaseModel):
    name: str
    config_url: str        
    url: str    

@app.get("/")
async def read_root():
    return "OK"


@app.post("/synthesize")
async def synthesize_text(request: TTSRequest):
    os.makedirs(AUDIO_OUTPUT_DIR, exist_ok=True)
    id = str(uuid.uuid4())
    text_filename = AUDIO_OUTPUT_DIR/f'{id}.txt'
    container_audio_path = AUDIO_OUTPUT_DIR/f"{id}.wav"
    try:
        model_path = os.path.join(MODEL_DIR, request.model + ".onnx")
        #TODO:
        #It works and it works fast enough, but how? Does it keep the model in memory for the next calls?
        #If not, can we somehow preload this model to make it even faster?
        with open(text_filename,'w') as stream:
            stream.write(request.text)
        command = [f'cat {text_filename} | /lsiopy/bin/piper --model {model_path} --output_file {container_audio_path}']
        for option in ['noise_scale', 'length_scale', 'noise_w','speaker']:
            value = getattr(request, option)
            if value is not None:
                command.append(f' --{option} {value}')
        command = ' '.join(command)

        try:
            subprocess.check_output(["sh", "-c", command])
        except subprocess.CalledProcessError as ex:
            raise ValueError(f"Exception when called\n{command}\n\nThe error: {ex.output}")

        return FileResponse(
            path=container_audio_path,
            media_type="audio/wav",
            filename=container_audio_path.name
        )
    except:
        raise HTTPException(status_code=500, detail=traceback.format_exc())
    finally:
        pass


import json
from typing import Iterable
from unittest import TestCase

from ....framework import (
    File, RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder,
    DockerWebServiceController, BrainBoxApi, BrainBoxTask, IModelDownloadingController, DownloadableModel
)
from .settings import FasterWhisperSettings, FasterWhisperModel
from pathlib import Path


class FasterWhisperController(DockerWebServiceController[FasterWhisperSettings], IModelDownloadingController):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return FasterWhisperModel

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            None,
            publish_ports={self.connection_settings.port:8084},
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return FasterWhisperSettings()

    def create_api(self):
        from .api import FasterWhisper
        return FasterWhisper()

    def run_notebook(self):
        self.run_with_configuration(self.get_service_run_configuration(None).as_notebook_service())

    def post_install(self):
        self.download_models(self.settings.models_to_download)


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import FasterWhisper

        file = File.read(Path(__file__).parent / 'files/test_voice.wav')

        first_time = True
        for model in self.settings.models_to_download:
            result = api.execute(BrainBoxTask.call(FasterWhisper).transcribe_json(file, model.name))
            yield TestReport.last_call(api).with_comment("Speech recognition with FasterWhisper, full output")
            tc.assertEqual(
                'One little spark and before you know it, the whole world is burning.',
                result['text'].strip()
            )

            result = api.execute(BrainBoxTask.call(FasterWhisper).transcribe(file, model.name))
            if first_time:
                yield TestReport.last_call(api).href('recognition').with_comment("Speech recognition with FasterWhisper, text-only output")
            tc.assertEqual(
                'One little spark and before you know it, the whole world is burning.',
                result
            )



DOCKERFILE=f'''
FROM python:3.9

{SmallImageBuilder.APT_INSTALL('ffmpeg')}

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

COPY . /home/app/

RUN pip freeze

ENTRYPOINT ["python3","/home/app/main.py"]

'''


DEPENDENCIES = '''
flask
notebook
faster-whisper
tqdm
ipywidgets
jupyter
hf_xet
'''

ORIGINAL_DEPENDENCIES='''
flask
notebook
faster-whisper
tqdm
ipywidgets
jupyter
hf_xet
'''


from pathlib import Path
from typing import cast
from ....framework import DownloadableModel, IController, FileIO
from dataclasses import dataclass
import requests

@dataclass
class HelloBrainBoxModel(DownloadableModel):
    name: str
    url: str

    def get_local_location(self, controller: IController) -> Path:
        return controller.resource_folder('models')/self.name

    @classmethod
    def download(cls, models: list['DownloadableModel'], controller: IController):
        for model in models:
            model = cast(HelloBrainBoxModel, model)
            reply = requests.get(model.url)
            FileIO.write_bytes(reply.content, model.get_local_location(controller))



from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, INotebookableController, BrainboxImageBuilder
)
from .settings import CoquiTTSSettings
from pathlib import Path


class CoquiTTSController(DockerWebServiceController[CoquiTTSSettings], INotebookableController):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.11.11',
            ('espeak-ng',),
            BrainboxImageBuilder.Repository(
                'https://github.com/coqui-ai/TTS',
                'dbf1a08a0d4e47fdad6172e433eeb34bc6b13b4e',
                '[all,dev,notebooks]'
            ),
            allow_arm64=False
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port: 8084},
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return CoquiTTSSettings()

    def create_api(self):
        from .api import CoquiTTS
        return CoquiTTS()

    def post_install(self):
        for model in self.settings.builtin_models_to_download:
            path = self.resource_folder('builtin') / model.due_folder_name
            if not path.is_dir():
                cfg = self.get_service_run_configuration(None).as_service_worker(
                    '--install',
                    model.model_name,
                    '--install-mode',
                    model.mode
                )
                self.run_auxiliary_configuration(cfg)

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .tests import Test
        yield TestReport.attach_source_file(Test)
        yield from Test(api, tc).test_all(self.settings)








from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, FileIO, INotebookableController, IModelDownloadingController, DownloadableModel,
    File
)
from .settings import ResembleEnhanceSettings
from ...common import check_if_its_sound
from pathlib import Path


class ResembleEnhanceController(
    DockerWebServiceController[ResembleEnhanceSettings],
    INotebookableController,
):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.11.11',
            ('git-lfs','libsndfile1', 'libsndfile1-dev'),
            allow_arm64=True,
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            publish_ports={self.connection_settings.port:8080},
            mount_resource_folders={
                'uploads' : '/uploads',
                'outputs' : '/outputs',
                'models' : '/home/app/.local/lib/python3.11/site-packages/resemble_enhance/model_repo'
            },
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()



    def get_default_settings(self):
        return ResembleEnhanceSettings()

    def create_api(self):
        from .api import ResembleEnhance
        return ResembleEnhance()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import ResembleEnhance

        input = File.read(Path(__file__).parent/'sample_10db.wav')
        result = api.execute(BrainBoxTask.call(ResembleEnhance).process(input))
        for file in result:
            check_if_its_sound(api.open_file(file).content, tc)

        yield (
            TestReport
            .last_call(api)
            .href('10db')
            .result_is_array_of_files(File.Kind.Audio)
        )



from typing import Iterable
from abc import ABC, abstractmethod
from ..app import BrainBoxApi
from ...common import FileLike, File
from pathlib import Path
from dataclasses import dataclass

class IPrerequisite(ABC):
    @abstractmethod
    def execute(self, api: BrainBoxApi):
        pass

@dataclass
class ResourcePrerequisite(IPrerequisite):
    decider: type|str
    resource_path: str
    content: bytes|Path|File|None = None

    def execute(self, api: BrainBoxApi):
        if self.content is None:
            api.controller_api.delete_resource(self.decider, self.resource_path, True)
        else:
            api.controller_api.upload_resource(self.decider, self.resource_path, self.content)


@dataclass
class CacheUploadPrerequisite(IPrerequisite):
    content: FileLike.Type
    file_name: str|None = None

    def execute(self, api: BrainBoxApi):
        filename = self.file_name
        if filename is None:
            filename = FileLike.get_name(self.content, True)
        api.upload(filename, self.content)


class CombinedPrerequisite(IPrerequisite):
    def __init__(self, prereq: Iterable[IPrerequisite]):
        self.prereq = tuple(prereq)

    def execute(self, api: BrainBoxApi):
        for prereq in self.prereq:
            prereq.execute(api)


from .architecture import AbstractRepresentationStep, RepresentationData
from dataclasses import dataclass

@dataclass
class SortedRepresentationStep(AbstractRepresentationStep):
    def select(self, data: list[RepresentationData]) -> None:
        data = list(sorted(data, key = lambda z: z.category_size))
        for i in range(self.count):
            data[i].selected = True



from .docker_controller import DockerController, TSettings
from typing import Generic
from abc import ABC, abstractmethod
from .on_demand_docker_api import OnDemandDockerApi

class OnDemandDockerController(DockerController[TSettings]):
    @abstractmethod
    def create_api(self) -> OnDemandDockerApi:
        pass

    def find_api(self, instance_id: str):
        api = self.create_api()
        api.controller = self
        return api

    def run(self, parameter: str|None):
        return 'no-instance'

    def stop(self, instance_id: str):
        self.get_deployment().stop(True)







import os
import sys
import subprocess
from pathlib import Path
import argparse
import shutil

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-f','--fix', action='store_true')
    parser.add_argument('-i','--install')
    parser.add_argument('-c','--commit')

    args = parser.parse_args()
    print(f"Running with arguments\n{args}")

    if args.fix:
        fix_path = Path('/home/app/fix')
        comfy_path = Path('/home/app/ComfyUI')
        for folder_name in ['models', 'input']:
            for item in os.listdir(f'/home/app/fix/{folder_name}'):
                src = fix_path/folder_name/item
                dst = Path()/folder_name/item
                if not dst.is_file() and not dst.is_dir():
                    if src.is_dir():
                        shutil.copytree(src, dst)
                    elif src.is_file():
                        os.makedirs(dst.parent, exist_ok=True)
                        shutil.copy(src, dst)
        exit(0)

    if args.install:
        subprocess.call(
            ['git', 'clone', args.install],
            cwd='/home/app/ComfyUI/custom_nodes'
        )
        if args.commit is not None:
            subprocess.call(
                ['git','reset','--hard', args.commit]
            )
        exit(0)

    subprocess.run([sys.executable, '/home/app/ComfyUI/main.py', '--listen'])




from .step import IStep


class ConstantStep(IStep):
    def __init__(self, records: list):
        self.records = records

    def process(self, history, current):
        return self.records


class FilterStep(IStep):
    def __init__(self, filter):
        self.filter = filter

    def process(self, history, current):
        return [o for o in current if self.filter(o)]




from brainbox.deciders import BoilerplateServer
from brainbox.utils.compatibility import *

if __name__ == '__main__':
    controller = BoilerplateServer.Controller()
    #resolve_dependencies(controller)

    controller.install()
    controller.self_test()

    #test_on_arm64(controller)



from brainbox.deciders import WhisperX
from brainbox.framework import Loc
import dotenv
from brainbox.utils.compatibility import *

if __name__ == '__main__':
    dotenv.load_dotenv(Loc.root_folder/'environment.env')
    controller = WhisperX.Controller()
    #resolve_dependencies(controller)

    controller.install()
    controller.self_test()


from brainbox.deciders import CoquiTTS
from brainbox.utils.compatibility import *

if __name__ == '__main__':
    installer = CoquiTTS.Controller()
    #resolve_dependencies(installer)

    installer.install()
    installer.self_test()
    #installer.run_notebook()






from brainbox.deciders import Resemblyzer
from brainbox.utils.compatibility import *

if __name__ == '__main__':
    controller = Resemblyzer.Controller()
    #resolve_dependencies(controller)
    controller.install()
    controller.self_test()
    #controller.run_notebook()
    # test_on_arm64(controller)





import flask
import shutil
import os
import pickle
import uuid
import traceback
from tagger.interrogators import interrogators
from pathlib import Path
from PIL import Image

class WD14Server:
    def __init__(self):
        self.interrogators = {}

    def __call__(self):
        app = flask.Flask(__name__)
        app.add_url_rule('/interrogate/<model>/<threshold>', view_func=self.interrogate, methods=['POST'])
        app.add_url_rule('/tags/<model>', view_func=self.tags, methods=['POST'])
        app.add_url_rule('/', view_func=self.index, methods=['GET'])
        app.run('0.0.0.0',8084)

    def index(self):
        return "WD14 is running"

    def _get_interrogator(self, model):
        if model not in self.interrogators:
            self.interrogators[model] = interrogators[model]
            self.interrogators[model].use_cpu()
            self.interrogators[model].load()
        return self.interrogators[model]

    def interrogate(self, model, threshold):
        try:
            threshold = float(threshold)
            file = flask.request.files['file']
            image = Image.open(file)
            interrogator = self._get_interrogator(model)
            result = interrogator.interrogate(image)
            result = result[1]
            output = {key: value for key, value in result.items() if value>=threshold}
            return flask.jsonify(output)
        except:
            return traceback.format_exc(), 500


    def tags(self, model):
        try:
            data = flask.request.json
            count = data.get('count', None)
            interrogator = self._get_interrogator(model)
            tags = list([row[1].to_dict() for row in interrogator.tags.iterrows()])
            if count is not None:
                tags = tags[:count]
            return flask.jsonify(tags)
        except:
            return traceback.format_exc(), 500











from dataclasses import dataclass
from .builder_part import IBuilderPart, BuildContext

@dataclass
class AptInstallation(IBuilderPart):
    packages: tuple[str,...]

    def to_commands(self, context: BuildContext) -> list[str]:
        libs = ' '.join(self.packages)
        return [(
            'RUN apt-get update && '
            f'apt-get install -y --no-install-recommends {libs} && '
            'rm -rf /var/lib/apt/lists/*'
        )]


from dataclasses import dataclass
from .build_context import BuildContext

@dataclass
class Repository:
    url: str
    commit: str|None = None
    options: str = ''
    install:bool = True
    remove_files: tuple[str,...]|None = None

    def to_commands(self, context: BuildContext) -> list[str]:
        result = []

        line_init = []
        line_init.append(f'RUN git clone {self.url} /home/app/repo')
        line_init.append(f' && cd /home/app/repo')
        if self.commit is not None:
            line_init.append(f' && git reset --hard {self.commit}')
        line_init.append(' && rm -rf .git')

        result.append(''.join(line_init))

        if context.repo_fix_folder.is_dir():
            result.append(f'COPY {context.repo_fix_folder.name}/ /home/app/repo/')

        if self.remove_files is not None:
            line_remove = ['RUN ']
            for index, file in enumerate(self.remove_files):
                if index>0:
                    line_remove.append(' && ')
                line_remove.append('rm /home/app/repo/'+file)
            result.append(''.join(line_remove))

        if self.install:
            line_install = []
            line_install.append(f'RUN cd /home/app/repo && pip install --user -e .{self.options}')
            line_install.append(' && rm -rf ~/.cache/pip')
            result.append(''.join(line_install))

        return result


from dataclasses import dataclass
from enum import Enum
from .controller_context import ControllerContext
from .resource_folder import ResourceFolder
from ...common import Loc
from ...deployment import DockerArgumentsHelper, LocalExecutor, Command
from copy import deepcopy
from .gpu_registry import GpuRegistry




@dataclass
class RunConfiguration:
    parameter: str|None = None
    publish_ports: None | dict[int, int] = None
    mount_resource_folders: None | dict[str, str] = None
    mount_top_resource_folder: bool = True
    propagate_env_variables: None | list[str] = None
    set_env_variables: None | dict[str, str] = None
    command_line_arguments: list[str] | tuple[str, ...] = ()
    custom_flags: tuple[str, ...] | list[str] = ()
    dont_rm: bool = False
    restart_unless_stopped: bool = False
    run_as_root: bool = False
    mount_root_folder: bool = False
    mount_custom_folders: None | dict[str, str] = None

    detach_and_interactive: bool = True
    platform: str|None = None

    def _mounts(self, resource_folder: ResourceFolder):
        if self.mount_resource_folders is not None:
            mount_folders = {
                resource_folder(host_folder) : container_folder
                for host_folder, container_folder in self.mount_resource_folders.items()
            }
        else:
            mount_folders = {}
        if self.mount_top_resource_folder:
            mount_folders[resource_folder()] = '/resources'
        if self.mount_root_folder:
            mount_folders[Loc.root_folder] = '/repo'
        if self.mount_custom_folders is not None:
            for key, value in self.mount_custom_folders.items():
                mount_folders[key] = value
        return mount_folders

    def generate_command(self,
                         container_name: str,
                         image_name: str,
                         context: ControllerContext,
                         ):
        arguments = ['docker', 'run']
        if self.publish_ports is not None:
            arguments += DockerArgumentsHelper.arg_publish_ports(self.publish_ports)
        arguments += DockerArgumentsHelper.arg_mount_folders(self._mounts(context.resource_folder))
        arguments += DockerArgumentsHelper.arg_propagate_env_variables(self.propagate_env_variables)
        arguments += DockerArgumentsHelper.arg_set_env_variables(self.set_env_variables)
        if GpuRegistry.are_present():
            arguments += ['--gpus', 'all']
        arguments += ['--interactive', '--tty', '--detach'] if self.detach_and_interactive else []
        arguments += ['--rm'] if not self.dont_rm else []
        arguments += ['--user',f'{context.machine.user_id}:{context.machine.group_id}'] if not self.run_as_root else []
        arguments += ['--restart', 'unless-stopped'] if self.restart_unless_stopped else []
        arguments += list(self.custom_flags)
        arguments += ['--name', container_name]
        arguments += ['--env', f'BRAINBOX_PARAMETER='+('*'+self.parameter if self.parameter is not None else '')]
        if self.platform is not None:
            arguments+=['--platform', self.platform]
        arguments += [image_name]
        arguments += list(self.command_line_arguments)

        return arguments

    def as_notebook_service(self):
        configuration = deepcopy(self)
        configuration.command_line_arguments=['--notebook']
        configuration.publish_ports = {8899:8899}
        configuration.mount_data_folder = True
        configuration.mount_custom_folders = {Loc.root_folder: '/repo'}
        configuration.detach_and_interactive = False
        return configuration

    def as_service_worker(self, *arguments):
        configuration = deepcopy(self)
        configuration.command_line_arguments = list(arguments)
        configuration.publish_ports = {}
        configuration.mount_top_resource_folder = True
        configuration.detach_and_interactive = False
        return configuration



from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, INotebookableController, IModelDownloadingController, File, DownloadableModel
)
from .settings import PiperSettings
from pathlib import Path
from ...common import VOICEOVER_TEXT, check_if_its_sound
from .model import PiperModel

class PiperController(
    DockerWebServiceController[PiperSettings],
    IModelDownloadingController
):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            'lscr.io/linuxserver/piper:1.6.3',
            ('python3', 'python3-pip', 'wget'),
            install_requirements_as_root=True,
            allow_arm64=True
        )

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return PiperModel

    def post_install(self):
        self.download_models(self.settings.models_to_download)

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            publish_ports={self.connection_settings.port:8080},
            mount_resource_folders={
                'models' : '/models',
                'cache' : '/cache'
            },
        )

    def get_default_settings(self):
        return PiperSettings

    def create_api(self):
        from .api import Piper
        return Piper()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import Piper

        task = BrainBoxTask.call(Piper).voiceover(
            text=VOICEOVER_TEXT,
            model="en",
        )
        result_file = api.execute(task)
        yield TestReport.last_call(api).result_is_file(File.Kind.Audio).href('href')
        check_if_its_sound(api.open_file(result_file).content, tc)







import os
from typing import Iterable
from unittest import TestCase
from brainbox.framework import Loc, FileIO, DownloadableModel
from .model import PiperTrainingModel
from ...common import VOICEOVER_TEXT, check_if_its_sound

from ....framework import (
    TestReport, SmallImageBuilder, IImageBuilder,
    BrainBoxApi, BrainBoxTask, File,
    OnDemandDockerController, IModelDownloadingController
)
from .settings import PiperTrainingSettings
from pathlib import Path
import zipfile



class PiperTrainingController(OnDemandDockerController[PiperTrainingSettings], IModelDownloadingController):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_default_settings(self):
        return PiperTrainingSettings()

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return PiperTrainingModel

    def create_api(self):
        from .api import PiperTraining
        return PiperTraining()

    def post_install(self):
        self.download_models(self.settings.models_to_download)

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import PiperTraining
        from ..piper import Piper

        NAME  = 'piper_voice_training_dataset'

        #preparing zip-file in a dataset format
        src_folder = Path(__file__).parent/'files'
        dataset = Loc.test_folder/(NAME+'.zip')
        with zipfile.ZipFile(dataset, 'w') as zip:
            for file in os.listdir(src_folder):
                for instance in range(100):
                    zip.write(src_folder/file, f'voice/{instance}-'+file)

        task = PiperTraining.create_training_pack(
            dataset,
            PiperTraining.TrainingSettings(),
            clean_up_afterwards=False,
            reset_if_existing=True
        )
        result = api.execute(task)
        yield TestReport.last_call(api).href('training')

        Piper.UploadVoice(api.cache_folder/result[0]['onnx'], custom_name=NAME).execute(api)
        voice_result = api.execute(BrainBoxTask.call(Piper).voiceover(VOICEOVER_TEXT, NAME))
        check_if_its_sound(api.open_file(voice_result).content, tc)
        yield TestReport.last_call(api).href('inference').result_is_file(File.Kind.Audio)

        Piper.DeleteVoice(NAME).execute(api)



DOCKERFILE = f"""
FROM python:3.9.21

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

WORKDIR /home/app

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

{SmallImageBuilder.GIT_CLONE_AND_RESET(
    "https://github.com/rhasspy/piper",
    "piper",
    install=False
)}

RUN rm /home/app/piper/src/python/requirements.txt

RUN sed -i 's/from . import attentions, commons, modules, monotonic_align/from . import attentions, commons, modules\\nimport monotonic_align/' /home/app/piper/src/python/piper_train/vits/models.py

COPY fixes/lightning.py /home/app/piper/src/python/piper_train/vits/lightning.py

COPY fixes/main.py /home/app/piper/src/python/piper_train/__main__.py

WORKDIR /home/app/piper/src/python

RUN pip install --user -e .

COPY . /home/app

ENTRYPOINT ["python3", "/home/app/main.py"]
"""

DEPENDENCIES = """
aiohappyeyeballs==2.4.6
aiohttp==3.11.12
aiosignal==1.3.2
async-timeout==5.0.1
attrs==25.1.0
audioread==3.0.1
certifi==2025.1.31
cffi==1.17.1
charset-normalizer==3.4.1
coloredlogs==15.0.1
Cython==0.29.37
decorator==5.1.1
flatbuffers==25.2.10
frozenlist==1.5.0
fsspec==2025.2.0
humanfriendly==10.0
idna==3.10
joblib==1.4.2
lazy_loader==0.4
librosa==0.10.2.post1
lightning-utilities==0.12.0
llvmlite==0.43.0
monotonic-align==1.0.0
mpmath==1.3.0
msgpack==1.1.0
multidict==6.1.0
numba==0.60.0
numpy==1.26.4
nvidia-cublas-cu11==11.10.3.66
nvidia-cuda-nvrtc-cu11==11.7.99
nvidia-cuda-runtime-cu11==11.7.99
nvidia-cudnn-cu11==8.5.0.96
onnx==1.17.0
onnxruntime==1.19.2
packaging==24.2
piper-phonemize==1.1.0
platformdirs==4.3.6
pooch==1.8.2
propcache==0.2.1
protobuf==5.29.3
pycparser==2.22
pytorch-lightning==1.9.5
PyYAML==6.0.2
requests==2.32.3
scikit-learn==1.6.1
scipy==1.13.1
soundfile==0.13.1
soxr==0.5.0.post1
sympy==1.13.3
threadpoolctl==3.5.0
torch==1.13.1
torchmetrics==1.5.2
tqdm==4.67.1
typing_extensions==4.12.2
urllib3==2.3.0
yarl==1.18.3
"""

#This recepy worked https://github.com/rhasspy/piper/issues/295 (up to lightning update)